'''
Alan Perception Module.
Authors: Jeff, Jacky
'''
import logging

from .version import __version__
from .camera_intrinsics import CameraIntrinsics
from .orthographic_intrinsics import OrthographicIntrinsics
from .exceptions import SensorUnresponsiveException

try:
    from .cnn import AlexNet, AlexNetWeights, conv
    from .features import Feature, LocalFeature, GlobalFeature, SHOTFeature, MVCNNFeature, BagOfFeatures
    from .feature_extractors import FeatureExtractor, CNNBatchFeatureExtractor, CNNReusableBatchFeatureExtractor
except Exception:
    logging.warning('Unable to import CNN modules! Likely due to missing tensorflow.')
    logging.warning('TensorFlow can be installed following the instructions in https://www.tensorflow.org/get_started/os_setup')

from .feature_matcher import Correspondences, NormalCorrespondences, FeatureMatcher, RawDistanceFeatureMatcher, PointToPlaneFeatureMatcher
from .image import Image, ColorImage, DepthImage, IrImage, GrayscaleImage, RgbdImage, GdImage, SegmentationImage, BinaryImage, PointCloudImage, NormalCloudImage
from .object_render import RenderMode, ObjectRender, QueryImageBundle
from .chessboard_registration import ChessboardRegistrationResult, CameraChessboardRegistration
from .point_registration import RegistrationResult, IterativeRegistrationSolver, PointToPlaneICPSolver
from .detector import RgbdDetection, RgbdDetector, RgbdForegroundMaskDetector, RgbdForegroundMaskQueryImageDetector, PointCloudBoxDetector, RgbdDetectorFactory
from .camera_sensor import CameraSensor, VirtualSensor, TensorDatasetVirtualSensor
from .webcam_sensor import WebcamSensor

try:
    from .kinect2_sensor import Kinect2PacketPipelineMode, Kinect2FrameMode, Kinect2RegistrationMode, Kinect2DepthMode, Kinect2BridgedQuality, Kinect2Sensor, KinectSensorBridged, VirtualKinect2Sensor, Kinect2SensorFactory, load_images
except Exception:
    logging.warning('Unable to import Kinect2 sensor modules! Likely due to missing pylibfreenect2.')
    logging.warning('The pylibfreenect2 library can be installed from https://github.com/r9y9/pylibfreenect2')

try:
    from .primesense_sensor import PrimesenseSensor, PrimesenseSensor_ROS, PrimesenseRegistrationMode
except Exception:
    logging.warning('Unable to import Primsense sensor modules! Likely due to missing OpenNI2.')

try:
    from .realsense_sensor import RealSenseSensor
except Exception:
    logging.warning('Unable to import RealSense sensor modules!')

try:
    from .ensenso_sensor import EnsensoSensor
except Exception:
    logging.warning('Unable to import Ensenso sensor modules!.')

try:
    from .phoxi_sensor import PhoXiSensor
    from .colorized_phoxi_sensor import ColorizedPhoXiSensor
except Exception as e:
    logging.warning('Unable to import PhoXi sensor modules!')

try:
    from .opencv_camera_sensor import OpenCVCameraSensor
    from .rgbd_sensors import RgbdSensorFactory
except Exception:
    logging.warning('Unable to import generic sensor modules!.')

try:
    from .weight_sensor import WeightSensor
except:
    logging.warning('Unable to import weight sensor modules!')

from .video_recorder import VideoRecorder

__all__ = [
    'CameraIntrinsics',
    'AlexNetWeights', 'AlexNet', 'conv',
    'RgbdDetection', 'RgbdDetector', 'RgbdForegroundMaskDetector', 'RgbdForegroundMaskQueryImageDetector', 'PointCloudBoxDetector', 'RgbdDetectorFactory',
    'FeatureExtractor', 'CNNBatchFeatureExtractor', 'CNNReusableBatchFeatureExtractor',
    'Correspondences', 'NormalCorrespondences', 'FeatureMatcher', 'RawDistanceFeatureMatcher', 'PointToPlaneFeatureMatcher',
    'Feature', 'LocalFeature', 'GlobalFeature', 'SHOTFeature', 'MVCNNFeature', 'BagOfFeatures',
    'Image', 'ColorImage', 'DepthImage', 'IrImage', 'GrayscaleImage', 'RgbdImage', 'GdImage', 'SegmentationImage', 'BinaryImage', 'PointCloudImage', 'NormalCloudImage',
    'Kinect2PacketPipelineMode', 'Kinect2FrameMode', 'Kinect2RegistrationMode', 'Kinect2DepthMode', 'Kinect2BridgedQuality', 'Kinect2Sensor','KinectSensorBridged','VirtualKinect2Sensor', 'Kinect2SensorFactory', 'load_images',
    'EnsensoSensor',
    'RgbdSensorFactory', 'PrimesenseSensor', 'VirtualPrimesenseSensor', 'PrimesenseSensor_ROS', 'PrimesenseRegistrationMode',
    'RealSenseSensor',
    'RenderMode', 'ObjectRender', 'QueryImageBundle',
    'RegistrationResult', 'IterativeRegistrationSolver', 'PointToPlaneICPSolver',
    'OpenCVCameraSensor',
    'VideoRecorder',
]



"""
Classes for feature matching between point sets for registration
Author: Jeff Mahler
"""
from abc import ABCMeta, abstractmethod
import numpy as np
from scipy import spatial
import scipy.spatial.distance as ssd

from autolab_core import PointCloud

from .features import BagOfFeatures

class Correspondences:
    """ Wrapper for point-set correspondences.

    Attributes
    ----------
    index_map : :obj:`list` of int
        maps list indices (source points) to target point indices
    source_points : Nx3 :obj:`numpy.ndarray`
        set of source points for registration
    target_points : Nx3 :obj:`numpy.ndarray`
        set of target points for registration
    num_matches : int
        the total number of matches
    """
    def __init__(self, index_map, source_points, target_points):
        self.index_map_ = index_map
        self.source_points_ = source_points
        self.target_points_ = target_points
        self.num_matches_ = source_points.shape[0]

    @property
    def index_map(self):
        return self.index_map_

    @property
    def source_points(self):
        return self.source_points_

    @property
    def target_points(self):
        return self.target_points_

    @property
    def num_matches(self):
        return self.num_matches_

    # Functions to iterate through matches like "for source_corr, target_corr in correspondences"
    def __iter__(self):
        self.iter_count_ = 0

    def next(self):
        if self.iter_count_ >= len(self.num_matches_):
            raise StopIteration
        else:
            return self.source_points_[self.iter_count,:], self.target_points_[self.iter_count,:]

class NormalCorrespondences(Correspondences):
    """ Wrapper for point-set correspondences with surface normals.

    Attributes
    ----------
    index_map : :obj:`list` of int
        maps list indices (source points) to target point indices
    source_points : Nx3 :obj:`numpy.ndarray`
        set of source points for registration
    target_points : Nx3 :obj:`numpy.ndarray`
        set of target points for registration
    source_normals : normalized Nx3 :obj:`numpy.ndarray`
        set of source normals for registration
    target_normals : normalized Nx3 :obj:`numpy.ndarray`
        set of target points for registration
    num_matches : int
        the total number of matches
    """
    def __init__(self, index_map, source_points, target_points, source_normals, target_normals):
        self.source_normals_ = source_normals
        self.target_normals_ = target_normals
        Correspondences.__init__(self, index_map, source_points, target_points)

    @property
    def source_normals(self):
        return self.source_normals_

    @property
    def target_normals(self):
        return self.target_normals_

    # Functions to iterate through matches like "for source_corr, target_corr in correspondences"
    def __iter__(self):
        self.iter_count_ = 0

    def next(self):
        if self.iter_count_ >= len(self.num_matches_):
            raise StopIteration
        else:
            return self.source_points_[self.iter_count,:], self.target_points_[self.iter_count,:], self.source_normals_[self.iter_count,:], self.target_normals_[self.iter_count,:]

class FeatureMatcher:
    """
    Generic feature matching between local features on a source and target object using nearest neighbors.
    """
    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    @staticmethod
    def get_point_index(point, all_points, eps = 1e-4):
        """ Get the index of a point in an array """
        inds = np.where(np.linalg.norm(point - all_points, axis=1) < eps)
        if inds[0].shape[0] == 0:
            return -1
        return inds[0][0]

    @abstractmethod
    def match(self, source_obj, target_obj):
        """
        Matches features between a source and target object. Source and target object types depend on subclass implementation.
        """
        pass

class RawDistanceFeatureMatcher(FeatureMatcher):
    def match(self, source_obj_features, target_obj_features):
        """
        Matches features between two graspable objects based on a full distance matrix.

        Parameters
        ----------
        source_obj_features : :obj:`BagOfFeatures`
            bag of the source objects features
        target_obj_features : :obj:`BagOfFeatures`
            bag of the target objects features

        Returns
        -------
        corrs : :obj:`Correspondences`
            the correspondences between source and target
        """
        if not isinstance(source_obj_features, f.BagOfFeatures):
            raise ValueError('Must supply source bag of object features')
        if not isinstance(target_obj_features, f.BagOfFeatures):
            raise ValueError('Must supply target bag of object features')

        # source feature descriptors and keypoints
        source_descriptors = source_obj_features.descriptors
        target_descriptors = target_obj_features.descriptors
        source_keypoints = source_obj_features.keypoints
        target_keypoints = target_obj_features.keypoints

        #calculate distance between this model's descriptors and each of the other_model's descriptors
        dists = spatial.distance.cdist(source_descriptors, target_descriptors)

        #calculate the indices of the target_model that minimize the distance to the descriptors in this model
        source_closest_descriptors = dists.argmin(axis=1)
        target_closest_descriptors = dists.argmin(axis=0)
        match_indices = []
        source_matched_points = np.zeros((0,3))
        target_matched_points = np.zeros((0,3))

        #calculate which points/indices the closest descriptors correspond to
        for i, j in enumerate(source_closest_descriptors):
            # for now, only keep correspondences that are a 2-way match
            if target_closest_descriptors[j] == i:
                match_indices.append(j)
                source_matched_points = np.r_[source_matched_points, source_keypoints[i:i+1, :]]
                target_matched_points = np.r_[target_matched_points, target_keypoints[j:j+1, :]]
            else:
                match_indices.append(-1)

        return Correspondences(match_indices, source_matched_points, target_matched_points)

class PointToPlaneFeatureMatcher(FeatureMatcher):
    """ Match points using a point to plane criterion with thresholding.

    Attributes
    ----------
    dist_thresh : float
        threshold distance to consider a match valid
    norm_thresh : float
        threshold cosine distance alignment betwen normals to consider a match valid
    """
    def __init__(self, dist_thresh=0.05, norm_thresh=0.75):
        self.dist_thresh_ = dist_thresh
        self.norm_thresh_ = norm_thresh
        FeatureMatcher.__init__(self)

    def match(self, source_points, target_points, source_normals, target_normals):
        """
        Matches points between two point-normal sets. Uses the closest ip to choose matches, with distance for thresholding only.

        Parameters
        ----------
        source_point_cloud : Nx3 :obj:`numpy.ndarray`
            source object points
        target_point_cloud : Nx3 :obj:`numpy.ndarray`
            target object points
        source_normal_cloud : Nx3 :obj:`numpy.ndarray`
            source object outward-pointing normals
        target_normal_cloud : Nx3 :obj`numpy.ndarray`
            target object outward-pointing normals

        Returns
        -------
        :obj`Correspondences`
            the correspondences between source and target
        """
        # compute the distances and inner products between the point sets
        dists = ssd.cdist(source_points, target_points, 'euclidean')
        ip = source_normals.dot(target_normals.T) # abs because we don't have correct orientations
        source_ip = source_points.dot(target_normals.T)
        target_ip = target_points.dot(target_normals.T)
        target_ip = np.diag(target_ip)
        target_ip = np.tile(target_ip, [source_points.shape[0], 1])
        abs_diff = np.abs(source_ip - target_ip) # difference in inner products

        # mark invalid correspondences
        invalid_dists = np.where(dists > self.dist_thresh_)
        abs_diff[invalid_dists[0], invalid_dists[1]] = np.inf
        invalid_norms = np.where(ip < self.norm_thresh_)
        abs_diff[invalid_norms[0], invalid_norms[1]] = np.inf

        # choose the closest matches
        match_indices = np.argmin(abs_diff, axis=1)
        match_vals = np.min(abs_diff, axis=1)
        invalid_matches = np.where(match_vals == np.inf)
        match_indices[invalid_matches[0]] = -1

        return NormalCorrespondences(match_indices, source_points, target_points, source_normals, target_normals)



"""
Interface to the Ensenso N* Sensor
Author: Jeff Mahler
"""
import IPython
import logging
import numpy as np
import os
import struct
import sys
import time

try:
    from cv_bridge import CvBridge, CvBridgeError
    import rospy
    from sensor_msgs.msg import CameraInfo, PointCloud2
    import sensor_msgs.point_cloud2 as pc2
except ImportError:
    logging.warning("Failed to import ROS in ensenso_sensor.py. ROS functionality not available")
    
from .constants import MM_TO_METERS, INTR_EXTENSION
from . import CameraIntrinsics, CameraSensor, ColorImage, DepthImage, Image

class EnsensoSensor(CameraSensor):
    """ Class for interfacing with an Ensenso N* sensor.
    """
    def __init__(self, frame='ensenso'):
        # set member vars
        self._frame = frame
        self._initialized = False
        self._format = None
        self._camera_intr = None
        self._cur_depth_im = None
        self._running = False
        
    def __del__(self):
        """Automatically stop the sensor for safety.
        """
        if self.is_running:
            self.stop()
        
    def _set_format(self, msg):
        """ Set the buffer formatting. """
        num_points = msg.height * msg.width
        self._format = '<' + num_points * 'ffff'
            
    def _set_camera_properties(self, msg):
        """ Set the camera intrinsics from an info msg. """
        focal_x = msg.K[0]
        focal_y = msg.K[4]
        center_x = msg.K[2]
        center_y = msg.K[5]
        im_height = msg.height
        im_width = msg.width
        self._camera_intr = CameraIntrinsics(self._frame, focal_x, focal_y,
                                             center_x, center_y,
                                             height=im_height,
                                             width=im_width)

    def _depth_im_from_pointcloud(self, msg):
        """ Convert a pointcloud2 message to a depth image. """
        # set format
        if self._format is None:
            self._set_format(msg)

        # rescale camera intr in case binning is turned on
        if msg.height != self._camera_intr.height:
            rescale_factor = float(msg.height) / self._camera_intr.height
            self._camera_intr = self._camera_intr.resize(rescale_factor)
            
        # read num points
        num_points = msg.height * msg.width
            
        # read buffer
        raw_tup = struct.Struct(self._format).unpack_from(msg.data, 0)
        raw_arr = np.array(raw_tup)

        # subsample depth values and reshape
        depth_ind = 2 + 4 * np.arange(num_points)
        depth_buf = raw_arr[depth_ind]
        depth_arr = depth_buf.reshape(msg.height, msg.width)
        depth_im = DepthImage(depth_arr, frame=self._frame)

        return depth_im

    def _pointcloud_callback(self, msg):
        """ Callback for handling point clouds. """
        self._cur_depth_im = self._depth_im_from_pointcloud(msg)
        
    def _camera_info_callback(self, msg):
        """ Callback for reading camera info. """
        self._camera_info_sub.unregister()
        self._set_camera_properties(msg)
        
    @property
    def ir_intrinsics(self):
        """:obj:`CameraIntrinsics` : The camera intrinsics for the Ensenso IR camera.
        """
        return self._camera_intr

    @property
    def is_running(self):
        """bool : True if the stream is running, or false otherwise.
        """
        return self._running

    @property
    def frame(self):
        """:obj:`str` : The reference frame of the sensor.
        """
        return self._frame

    def start(self):
        """ Start the sensor """
        # initialize subscribers
        self._pointcloud_sub = rospy.Subscriber('/%s/depth/points' %(self.frame), PointCloud2, self._pointcloud_callback)
        self._camera_info_sub = rospy.Subscriber('/%s/left/camera_info' %(self.frame), CameraInfo, self._camera_info_callback)

        while self._camera_intr is None:
            time.sleep(0.1)
        
        self._running = True

    def stop(self):
        """ Stop the sensor """
        # check that everything is running
        if not self._running:
            logging.warning('Ensenso not running. Aborting stop')
            return False

        # stop subs
        self._pointcloud_sub.unregister()
        self._camera_info_sub.unregister()
        self._running = False
        return True

    def frames(self):
        """Retrieve a new frame from the Ensenso and convert it to a ColorImage,
        a DepthImage, and an IrImage.

        Returns
        -------
        :obj:`tuple` of :obj:`ColorImage`, :obj:`DepthImage`, :obj:`IrImage`, :obj:`numpy.ndarray`
            The ColorImage, DepthImage, and IrImage of the current frame.

        Raises
        ------
        RuntimeError
            If the Ensenso stream is not running.
        """
        # wait for a new image
        while self._cur_depth_im is None:
            time.sleep(0.01)
            
        # read next image
        depth_im = self._cur_depth_im
        color_im = ColorImage(np.zeros([depth_im.height,
                                        depth_im.width,
                                        3]).astype(np.uint8), frame=self._frame)
        self._cur_depth_im = None
        return color_im, depth_im, None

    def median_depth_img(self, num_img=1, fill_depth=0.0):
        """Collect a series of depth images and return the median of the set.

        Parameters
        ----------
        num_img : int
            The number of consecutive frames to process.

        Returns
        -------
        :obj:`DepthImage`
            The median DepthImage collected from the frames.
        """
        depths = []

        for _ in range(num_img):
            _, depth, _ = self.frames()
            depths.append(depth)

        median_depth = Image.median_images(depths)
        median_depth.data[median_depth.data == 0.0] = fill_depth
        return median_depth
        



"""
Encapsulates camera intrinsic parameters for projecting / deprojecitng points
Author: Jeff Mahler
"""
import copy
import logging
import numpy as np
import json
import os

from autolab_core import Point, PointCloud, ImageCoords

from .constants import INTR_EXTENSION
from .image import DepthImage, PointCloudImage

class CameraIntrinsics(object):
    """A set of intrinsic parameters for a camera. This class is used to project
    and deproject points.
    """

    def __init__(self, frame, fx, fy=None, cx=0.0, cy=0.0, skew=0.0, height=None, width=None):
        """Initialize a CameraIntrinsics model.

        Parameters
        ----------
        frame : :obj:`str`
            The frame of reference for the point cloud.
        fx : float
            The x-axis focal length of the camera in pixels.
        fy : float
            The y-axis focal length of the camera in pixels.
        cx : float
            The x-axis optical center of the camera in pixels.
        cy : float
            The y-axis optical center of the camera in pixels.
        skew : float
            The skew of the camera in pixels.
        height : float
            The height of the camera image in pixels.
        width : float
            The width of the camera image in pixels
        """
        self._frame = frame
        self._fx = float(fx)
        self._fy = float(fy)
        self._cx = float(cx)
        self._cy = float(cy)
        self._skew = float(skew)
        self._height = int(height)
        self._width = int(width)

        # set focal, camera center automatically if under specified
        if fy is None:
            self._fy = fx

        # set camera projection matrix
        self._K = np.array([[self._fx, self._skew, self._cx],
                            [       0,   self._fy, self._cy],
                            [       0,          0,        1]])

    @property
    def frame(self):
        """:obj:`str` : The frame of reference for the point cloud.
        """
        return self._frame

    @property
    def fx(self):
        """float : The x-axis focal length of the camera in pixels.
        """
        return self._fx

    @property
    def fy(self):
        """float : The y-axis focal length of the camera in pixels.
        """
        return self._fy

    @property
    def cx(self):
        """float : The x-axis optical center of the camera in pixels.
        """
        return self._cx

    @cx.setter
    def cx(self, z):
        self._cx = z
        self._K = np.array([[self._fx, self._skew, self._cx],
                            [       0,   self._fy, self._cy],
                            [       0,          0,        1]])

    @property
    def cy(self):
        """float : The y-axis optical center of the camera in pixels.
        """
        return self._cy

    @cy.setter
    def cy(self, z):
        self._cy = z
        self._K = np.array([[self._fx, self._skew, self._cx],
                            [       0,   self._fy, self._cy],
                            [       0,          0,        1]])

    @property
    def skew(self):
        """float : The skew of the camera in pixels.
        """
        return self._skew

    @property
    def height(self):
        """float : The height of the camera image in pixels.
        """
        return self._height

    @property
    def width(self):
        """float : The width of the camera image in pixels
        """
        return self._width

    @property
    def proj_matrix(self):
        """:obj:`numpy.ndarray` : The 3x3 projection matrix for this camera.
        """
        return self._K

    @property
    def K(self):
        """:obj:`numpy.ndarray` : The 3x3 projection matrix for this camera.
        """
        return self._K

    @property
    def vec(self):
        """:obj:`numpy.ndarray` : Vector representation for this camera.
        """
        return np.r_[self.fx, self.fy, self.cx, self.cy, self.skew, self.height, self.width]
    
    @property
    def rosmsg(self):
        """:obj:`sensor_msgs.CamerInfo` : Returns ROS CamerInfo msg 
        """
        from sensor_msgs.msg import CameraInfo, RegionOfInterest
        from std_msgs.msg import Header

        msg_header = Header()
        msg_header.frame_id = self._frame

        msg_roi = RegionOfInterest()
        msg_roi.x_offset = 0
        msg_roi.y_offset = 0
        msg_roi.height = 0
        msg_roi.width = 0
        msg_roi.do_rectify = 0

        msg = CameraInfo()
        msg.header = msg_header
        msg.height = self._height
        msg.width = self._width
        msg.distortion_model = 'plumb_bob'
        msg.D = [0.0, 0.0, 0.0, 0.0, 0.0]
        msg.K = [self._fx, 0.0, self._cx, 0.0, self._fy, self._cy, 0.0, 0.0, 1.0]
        msg.R = [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0]
        msg.P = [self._fx, 0.0, self._cx, 0.0, 0.0, self._fx, self._cy, 0.0, 0.0, 0.0, 1.0, 0.0]
        msg.binning_x = 0
        msg.binning_y = 0
        msg.roi = msg_roi

        return msg

    @staticmethod
    def from_vec(vec, frame='unassigned'):
        return CameraIntrinsics(frame,
                                fx=vec[0],
                                fy=vec[1],
                                cx=vec[2],
                                cy=vec[3],
                                skew=vec[4],
                                height=vec[5],
                                width=vec[6])
    
    def crop(self, height, width, crop_ci, crop_cj):
        """ Convert to new camera intrinsics for crop of image from original camera.

        Parameters
        ----------
        height : int
            height of crop window
        width : int
            width of crop window
        crop_ci : int
            row of crop window center
        crop_cj : int
            col of crop window center

        Returns
        -------
        :obj:`CameraIntrinsics`
            camera intrinsics for cropped window
        """
        cx = self.cx + float(width-1)/2 - crop_cj
        cy = self.cy + float(height-1)/2 - crop_ci
        cropped_intrinsics = CameraIntrinsics(frame=self.frame,
                                              fx=self.fx,
                                              fy=self.fy,
                                              skew=self.skew,
                                              cx=cx, cy=cy,
                                              height=height,
                                              width=width)
        return cropped_intrinsics

    def resize(self, scale):
        """ Convert to new camera intrinsics with parameters for resized image.
        
        Parameters
        ----------
        scale : float
            the amount to rescale the intrinsics
        
        Returns
        -------
        :obj:`CameraIntrinsics`
            camera intrinsics for resized image        
        """
        center_x = float(self.width-1) / 2
        center_y = float(self.height-1) / 2
        orig_cx_diff = self.cx - center_x
        orig_cy_diff = self.cy - center_y
        height = scale * self.height
        width = scale * self.width
        scaled_center_x = float(width-1) / 2
        scaled_center_y = float(height-1) / 2
        fx = scale * self.fx
        fy = scale * self.fy
        skew = scale * self.skew
        cx = scaled_center_x + scale * orig_cx_diff
        cy = scaled_center_y + scale * orig_cy_diff
        scaled_intrinsics = CameraIntrinsics(frame=self.frame,
                                              fx=fx, fy=fy, skew=skew, cx=cx, cy=cy,
                                              height=height, width=width)
        return scaled_intrinsics

    def project(self, point_cloud, round_px=True):
        """Projects a point cloud onto the camera image plane.

        Parameters
        ----------
        point_cloud : :obj:`autolab_core.PointCloud` or :obj:`autolab_core.Point`
            A PointCloud or Point to project onto the camera image plane.

        round_px : bool
            If True, projections are rounded to the nearest pixel.

        Returns
        -------
        :obj:`autolab_core.ImageCoords` or :obj:`autolab_core.Point`
            A corresponding set of image coordinates representing the given
            PointCloud's projections onto the camera image plane. If the input
            was a single Point, returns a 2D Point in the camera plane.

        Raises
        ------
        ValueError
            If the input is not a PointCloud or Point in the same reference
            frame as the camera.
        """
        if not isinstance(point_cloud, PointCloud) and not (isinstance(point_cloud, Point) and point_cloud.dim == 3):
            raise ValueError('Must provide PointCloud or 3D Point object for projection')
        if point_cloud.frame != self._frame:
            raise ValueError('Cannot project points in frame %s into camera with frame %s' %(point_cloud.frame, self._frame))

        points_proj = self._K.dot(point_cloud.data)
        if len(points_proj.shape) == 1:
            points_proj = points_proj[:, np.newaxis]
        point_depths = np.tile(points_proj[2,:], [3, 1])
        points_proj = np.divide(points_proj, point_depths)
        if round_px:
            points_proj = np.round(points_proj)

        if isinstance(point_cloud, Point):
            return Point(data=points_proj[:2,:].astype(np.int16), frame=self._frame)
        return ImageCoords(data=points_proj[:2,:].astype(np.int16), frame=self._frame)

    def project_to_image(self, point_cloud, round_px=True):
        """Projects a point cloud onto the camera image plane and creates
        a depth image. Zero depth means no point projected into the camera
        at that pixel location (i.e. infinite depth).

        Parameters
        ----------
        point_cloud : :obj:`autolab_core.PointCloud` or :obj:`autolab_core.Point`
            A PointCloud or Point to project onto the camera image plane.

        round_px : bool
            If True, projections are rounded to the nearest pixel.

        Returns
        -------
        :obj:`DepthImage`
            A DepthImage generated from projecting the point cloud into the
            camera.

        Raises
        ------
        ValueError
            If the input is not a PointCloud or Point in the same reference
            frame as the camera.
        """
        if not isinstance(point_cloud, PointCloud) and not (isinstance(point_cloud, Point) and point_cloud.dim == 3):
            raise ValueError('Must provide PointCloud or 3D Point object for projection')
        if point_cloud.frame != self._frame:
            raise ValueError('Cannot project points in frame %s into camera with frame %s' %(point_cloud.frame, self._frame))

        points_proj = self._K.dot(point_cloud.data)
        if len(points_proj.shape) == 1:
            points_proj = points_proj[:, np.newaxis]
        point_depths = points_proj[2,:]
        point_z = np.tile(point_depths, [3, 1])
        points_proj = np.divide(points_proj, point_z)
        if round_px:
            points_proj = np.round(points_proj)
        points_proj = points_proj[:2,:].astype(np.int16)

        valid_ind = np.where((points_proj[0,:] >= 0) & \
                             (points_proj[1,:] >= 0) & \
                             (points_proj[0,:] < self.width) & \
                             (points_proj[1,:] < self.height))[0]

        depth_data = np.zeros([self.height, self.width])
        depth_data[points_proj[1,valid_ind], points_proj[0,valid_ind]] = point_depths[valid_ind]
        return DepthImage(depth_data, frame=self.frame)

    def deproject(self, depth_image):
        """Deprojects a DepthImage into a PointCloud.

        Parameters
        ----------
        depth_image : :obj:`DepthImage`
            The 2D depth image to projet into a point cloud.

        Returns
        -------
        :obj:`autolab_core.PointCloud`
            A 3D point cloud created from the depth image.

        Raises
        ------
        ValueError
            If depth_image is not a valid DepthImage in the same reference frame
            as the camera.
        """
        # check valid input
        if not isinstance(depth_image, DepthImage):
            raise ValueError('Must provide DepthImage object for projection')
        if depth_image.frame != self._frame:
            raise ValueError('Cannot deproject points in frame %s from camera with frame %s' %(depth_image.frame, self._frame))

        # create homogeneous pixels
        row_indices = np.arange(depth_image.height)
        col_indices = np.arange(depth_image.width)
        pixel_grid = np.meshgrid(col_indices, row_indices)
        pixels = np.c_[pixel_grid[0].flatten(), pixel_grid[1].flatten()].T
        pixels_homog = np.r_[pixels, np.ones([1, pixels.shape[1]])]
        depth_arr = np.tile(depth_image.data.flatten(), [3,1])

        # deproject
        points_3d = depth_arr * np.linalg.inv(self._K).dot(pixels_homog)
        return PointCloud(data=points_3d, frame=self._frame)

    def deproject_to_image(self, depth_image):
        """Deprojects a DepthImage into a PointCloudImage.

        Parameters
        ----------
        depth_image : :obj:`DepthImage`
            The 2D depth image to projet into a point cloud.

        Returns
        -------
        :obj:`PointCloudImage`
            A point cloud image created from the depth image.

        Raises
        ------
        ValueError
            If depth_image is not a valid DepthImage in the same reference frame
            as the camera.
        """
        point_cloud = self.deproject(depth_image)
        point_cloud_im_data = point_cloud.data.T.reshape(depth_image.height, depth_image.width, 3)
        return PointCloudImage(data=point_cloud_im_data,
                               frame=self._frame)

    def deproject_pixel(self, depth, pixel):
        """Deprojects a single pixel with a given depth into a 3D point.

        Parameters
        ----------
        depth : float
            The depth value at the given pixel location.

        pixel : :obj:`autolab_core.Point`
            A 2D point representing the pixel's location in the camera image.

        Returns
        -------
        :obj:`autolab_core.Point`
            The projected 3D point.

        Raises
        ------
        ValueError
            If pixel is not a valid 2D Point in the same reference frame
            as the camera.
        """
        if not isinstance(pixel, Point) and not pixel.dim == 2:
            raise ValueError('Must provide 2D Point object for pixel projection')
        if pixel.frame != self._frame:
            raise ValueError('Cannot deproject pixel in frame %s from camera with frame %s' %(pixel.frame, self._frame))

        point_3d = depth * np.linalg.inv(self._K).dot(np.r_[pixel.data, 1.0])
        return Point(data=point_3d, frame=self._frame)

    def save(self, filename):
        """Save the CameraIntrinsics object to a .intr file.

        Parameters
        ----------
        filename : :obj:`str`
            The .intr file to save the object to.

        Raises
        ------
        ValueError
            If filename does not have the .intr extension.
        """
        file_root, file_ext = os.path.splitext(filename)
        if file_ext.lower() != INTR_EXTENSION:
            raise ValueError('Extension %s not supported for CameraIntrinsics. Must be stored with extension %s' %(file_ext, INTR_EXTENSION))

        camera_intr_dict = copy.deepcopy(self.__dict__)
        camera_intr_dict['_K'] = 0 # can't save matrix
        f = open(filename, 'w')
        json.dump(camera_intr_dict, f)
        f.close()

    @staticmethod
    def load(filename):
        """Load a CameraIntrinsics object from a file.

        Parameters
        ----------
        filename : :obj:`str`
            The .intr file to load the object from.

        Returns
        -------
        :obj:`CameraIntrinsics`
            The CameraIntrinsics object loaded from the file.

        Raises
        ------
        ValueError
            If filename does not have the .intr extension.
        """
        file_root, file_ext = os.path.splitext(filename)
        if file_ext.lower() != INTR_EXTENSION:
            raise ValueError('Extension %s not supported for CameraIntrinsics. Must be stored with extension %s' %(file_ext, INTR_EXTENSION))

        f = open(filename, 'r')
        ci = json.load(f)
        f.close()
        return CameraIntrinsics(frame=ci['_frame'],
                                fx=ci['_fx'],
                                fy=ci['_fy'],
                                cx=ci['_cx'],
                                cy=ci['_cy'],
                                skew=ci['_skew'],
                                height=ci['_height'],
                                width=ci['_width'])



import cv2
import logging
import numpy as np
import os
import subprocess
import shlex

from . import CameraSensor, ColorImage, CameraIntrinsics

class WebcamSensor(CameraSensor):
    """Class for interfacing with a Logitech webcam sensor (720p).
    """

    def __init__(self, frame='webcam', device_id=0):
        """Initialize a Logitech webcam sensor.

        Parameters
        ----------
        frame : str
            A name for the frame in which RGB images are returned.
        device_id : int
            The device ID for the webcam (by default, zero).
        """
        self._frame = frame
        self._camera_intr = None
        self._device_id = device_id
        self._cap = None
        self._running = False
        self._adjust_exposure = True

        # Set up camera intrinsics for the sensor
        width, height = 1280, 960
        focal_x, focal_y = 1430., 1420.
        center_x, center_y = 1280/2, 960/2
        self._camera_intr = CameraIntrinsics(self._frame, focal_x, focal_y,
                                             center_x, center_y,
                                             height=height, width=width)

    def __del__(self):
        """Automatically stop the sensor for safety.
        """
        if self.is_running:
            self.stop()

    @property
    def color_intrinsics(self):
        """CameraIntrinsics : The camera intrinsics for the PhoXi Greyscale camera.
        """
        return self._camera_intr

    @property
    def is_running(self):
        """bool : True if the stream is running, or false otherwise.
        """
        return self._running

    @property
    def frame(self):
        """str : The reference frame of the sensor.
        """
        return self._frame

    @property
    def color_frame(self):
        """str : The reference frame of the sensor.
        """
        return self._frame

    def start(self):
        """Start the sensor.
        """
        self._cap = cv2.VideoCapture(self._device_id + cv2.CAP_V4L2)
        if not self._cap.isOpened():
            self._running = False
            self._cap.release()
            self._cap = None
            return False

        self._cap.set(cv2.CAP_PROP_FRAME_WIDTH, self._camera_intr.width)
        self._cap.set(cv2.CAP_PROP_FRAME_HEIGHT, self._camera_intr.height)
        self._running = True

        # Capture 5 frames to flush webcam sensor
        for _ in range(5):
            _ = self.frames()

        return True

    def stop(self):
        """Stop the sensor.
        """
        # Check that everything is running
        if not self._running:
            logging.warning('Webcam not running. Aborting stop')
            return False

        if self._cap:
            self._cap.release()
            self._cap = None
        self._running = False

        return True

    def frames(self, most_recent=False):
        """Retrieve a new frame from the PhoXi and convert it to a ColorImage,
        a DepthImage, and an IrImage.

        Parameters
        ----------
        most_recent: bool
            If true, the OpenCV buffer is emptied for the webcam before reading the most recent frame.

        Returns
        -------
        :obj:`tuple` of :obj:`ColorImage`, :obj:`DepthImage`, :obj:`IrImage`, :obj:`numpy.ndarray`
            The ColorImage, DepthImage, and IrImage of the current frame.
        """
        if most_recent:
            for i in xrange(4):
                self._cap.grab()
        for i in range(1):
            if self._adjust_exposure:
                try:
                    command = 'v4l2-ctl -d /dev/video{} -c exposure_auto=1 -c exposure_auto_priority=0 -c exposure_absolute=100 -c saturation=60 -c gain=140'.format(self._device_id)
                    FNULL = open(os.devnull, 'w')
                    subprocess.call(shlex.split(command), stdout=FNULL, stderr=subprocess.STDOUT)
                except:
                    pass
            ret, frame = self._cap.read()
        rgb_data = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

        return ColorImage(rgb_data, frame=self._frame), None, None



import time
import logging

import numpy as np

try:
    import rospy
    from cv_bridge import CvBridge
    from sensor_msgs.msg import Image as ImageMessage
    from std_srvs.srv import Empty
    from perception.srv import ConnectCamera, GetDeviceList, GetFrame, TriggerImage
except ImportError:
    logging.warning("Failed to import ROS in phoxi_sensor.py. PhoXiSensor functionality unavailable.")

from . import CameraSensor, DepthImage, ColorImage, GrayscaleImage, CameraIntrinsics, Image, SensorUnresponsiveException

class PhoXiSensor(CameraSensor):
    """Class for interfacing with a PhoXi Structured Light Sensor.
    """

    def __init__(self, frame='phoxi', device_name='2018-02-020-LC3', size='small'):
        """Initialize a PhoXi Sensor.

        Parameters
        ----------
        frame : str
            A name for the frame in which depth images, normal maps, and RGB images are returned.
        device_name : str
            The string name of the PhoXi device (SN listed on sticker on back sensor).
            Old PhoXi: 1703005
            New PhoXi: 2018-02-020-LC3
        size : str
            An indicator for which size of image is desired.
            Either 'large' (2064x1544) or 'small' (1032x772).
        """

        self._frame = frame
        self._device_name = str(device_name)
        self._camera_intr = None
        self._running = False
        self._bridge = CvBridge()

        self._cur_color_im = None
        self._cur_depth_im = None
        self._cur_normal_map = None

        # Set up camera intrinsics for the sensor
        width, height = 2064, 1544
        focal_x, focal_y = 2244., 2244.
        center_x, center_y = 1023., 768.
        if size == 'small':
            width = 1032
            height = 772
            focal_x = focal_x / 2
            focal_y = focal_y / 2
            center_x = center_x / 2
            center_y = center_y / 2

            if str(device_name) == '1703005':
                focal_x = focal_y = 1105.

        self._camera_intr = CameraIntrinsics(self._frame, focal_x, focal_y,
                                             center_x, center_y,
                                             height=height, width=width)

    def __del__(self):
        """Automatically stop the sensor for safety.
        """
        if self.is_running:
            self.stop()

    @property
    def color_intrinsics(self):
        """CameraIntrinsics : The camera intrinsics for the PhoXi Greyscale camera.
        """
        return self._camera_intr

    @property
    def ir_intrinsics(self):
        """CameraIntrinsics : The camera intrinsics for the PhoXi IR camera.
        """
        return self._camera_intr

    @property
    def is_running(self):
        """bool : True if the stream is running, or false otherwise.
        """
        return self._running

    @property
    def frame(self):
        """str : The reference frame of the sensor.
        """
        return self._frame

    @property
    def color_frame(self):
        """str : The reference frame of the sensor.
        """
        return self._frame

    @property
    def ir_frame(self):
        """str : The reference frame of the sensor.
        """
        return self._frame

    def start(self):
        """Start the sensor.
        """
        if rospy.get_name() == '/unnamed':
            raise ValueError('PhoXi sensor must be run inside a ros node!')

        # Connect to the cameras
        if not self._connect_to_sensor():
            self._running = False
            return False

        # Set up subscribers for camera data
        self._color_im_sub = rospy.Subscriber('/phoxi_camera/texture', ImageMessage, self._color_im_callback)
        self._depth_im_sub = rospy.Subscriber('/phoxi_camera/depth_map', ImageMessage, self._depth_im_callback)
        self._normal_map_sub = rospy.Subscriber('/phoxi_camera/normal_map', ImageMessage, self._normal_map_callback)

        self._running = True

        return True

    def stop(self):
        """Stop the sensor.
        """
        # Check that everything is running
        if not self._running:
            logging.warning('PhoXi not running. Aborting stop')
            return False

        # Stop the subscribers
        self._color_im_sub.unregister()
        self._depth_im_sub.unregister()
        self._normal_map_sub.unregister()

        # Disconnect from the camera
        rospy.ServiceProxy('phoxi_camera/disconnect_camera', Empty)()

        self._running = False

        return True

    def frames(self):
        """Retrieve a new frame from the PhoXi and convert it to a ColorImage,
        a DepthImage, and an IrImage.

        Returns
        -------
        :obj:`tuple` of :obj:`ColorImage`, :obj:`DepthImage`, :obj:`IrImage`, :obj:`numpy.ndarray`
            The ColorImage, DepthImage, and IrImage of the current frame.
        """
        # Run a software trigger
        times = []

        rospy.ServiceProxy('phoxi_camera/start_acquisition', Empty)()
        rospy.ServiceProxy('phoxi_camera/trigger_image', TriggerImage)()

        self._cur_color_im = None
        self._cur_depth_im = None
        self._cur_normal_map = None

        rospy.ServiceProxy('phoxi_camera/get_frame', GetFrame)(-1)

        max_time = 5.0
        time_waiting = 0.0
        while self._cur_color_im is None or self._cur_depth_im is None or self._cur_normal_map is None:
            time.sleep(0.05)
            time_waiting += 0.05
            if time_waiting > max_time:
                raise SensorUnresponsiveException('PhoXi sensor seems to be non-responsive')
        return self._cur_color_im, self._cur_depth_im, None

    def median_depth_img(self, num_img=1, fill_depth=0.0):
        """Collect a series of depth images and return the median of the set.

        Parameters
        ----------
        num_img : int
            The number of consecutive frames to process.

        Returns
        -------
        DepthImage
            The median DepthImage collected from the frames.
        """
        depths = []

        for _ in range(num_img):
            _, depth, _ = self.frames()
            depths.append(depth)

        median_depth = Image.median_images(depths)
        median_depth.data[median_depth.data == 0.0] = fill_depth
        return median_depth

    def _connect_to_sensor(self):
        """Connect to the sensor.
        """
        name = self._device_name
        try:
            # Check if device is actively in list
            rospy.wait_for_service('phoxi_camera/get_device_list')
            device_list = rospy.ServiceProxy('phoxi_camera/get_device_list', GetDeviceList)().out
            if not str(name) in device_list:
                logging.error('PhoXi sensor {} not in list of active devices'.format(name))
                return False

            success = rospy.ServiceProxy('phoxi_camera/connect_camera', ConnectCamera)(name).success
            if not success:
                logging.error('Could not connect to PhoXi sensor {}'.format(name))
                return False

            logging.debug('Connected to PhoXi Sensor {}'.format(name))
            return True

        except rospy.ServiceException as e:
            logging.error('Service call failed: {}'.format(e))
            return False

    def _color_im_callback(self, msg):
        """Callback for handling textures (greyscale images).
        """
        try:
            data = self._bridge.imgmsg_to_cv2(msg)
            if np.max(data) > 255.0:
                data = 255.0 * data / 1200.0 # Experimentally set value for white
            data = np.clip(data, 0., 255.0).astype(np.uint8)
            gsimage = GrayscaleImage(data, frame=self._frame)
            self._cur_color_im = gsimage.to_color()
        except:
            self._cur_color_im = None

    def _depth_im_callback(self, msg):
        """Callback for handling depth images.
        """
        try:
            self._cur_depth_im = DepthImage(self._bridge.imgmsg_to_cv2(msg) / 1000.0, frame=self._frame)
        except:
            self._cur_depth_im = None

    def _normal_map_callback(self, msg):
        """Callback for handling normal maps.
        """
        try:
            self._cur_normal_map = self._bridge.imgmsg_to_cv2(msg)
        except:
            self._cur_normal_map = None



"""
Classes for easy chessboard registration
Authors: Jeff Mahler and Jacky Liang
"""
import cv2
import logging
import math
import matplotlib.pyplot as plt
import numpy as np
import time

from autolab_core import PointCloud, RigidTransform, Point
from .image import DepthImage

class ChessboardRegistrationResult(object):
    """ Struct to encapsulate results of camera-to-chessboard registration.
    
    Attributes
    ----------
    T_camera_cb : :obj:`autolab_core.RigidTransform`
        transformation from camera to chessboard frame
    cb_points_camera : :obj:`autolab_core.PointCloud`
        3D locations of chessboard corners in the camera frame
    """
    def __init__(self, T_camera_cb, cb_points_camera):
        self.T_camera_cb = T_camera_cb
        self.cb_points_cam = cb_points_camera

class CameraChessboardRegistration:
    """
    Namespace for camera to chessboard registration functions.
    """

    @staticmethod
    def register(sensor, config):
        """
        Registers a camera to a chessboard.

        Parameters
        ----------
        sensor : :obj:`perception.RgbdSensor`
            the sensor to register
        config : :obj:`autolab_core.YamlConfig` or :obj:`dict`
            configuration file for registration

        Returns
        -------
        :obj:`ChessboardRegistrationResult`
            the result of registration

        Notes
        -----
        The config must have the parameters specified in the Other Parameters section.

        Other Parameters
        ----------------
        num_transform_avg : int
            the number of independent registrations to average together
        num_images : int
            the number of images to read for each independent registration
        corners_x : int
            the number of chessboard corners in the x-direction
        corners_y : int
            the number of chessboard corners in the y-direction
        color_image_rescale_factor : float
            amount to rescale the color image for detection (numbers around 4-8 are useful)
        vis : bool
            whether or not to visualize the registration
        """
        # read config
        num_transform_avg = config['num_transform_avg']
        num_images = config['num_images']
        sx = config['corners_x']
        sy = config['corners_y']
        point_order = config['point_order']
        color_image_rescale_factor = config['color_image_rescale_factor']
        flip_normal = config['flip_normal']
        y_points_left = False
        if 'y_points_left' in config.keys() and sx == sy:
            y_points_left = config['y_points_left']
            num_images = 1
        vis = config['vis']

        # read params from sensor
        logging.info('Registering camera %s' %(sensor.frame))
        ir_intrinsics = sensor.ir_intrinsics

        # repeat registration multiple times and average results
        R = np.zeros([3,3])
        t = np.zeros([3,1])
        points_3d_plane = PointCloud(np.zeros([3, sx*sy]), frame=sensor.ir_frame)

        k = 0
        while k < num_transform_avg:
            # average a bunch of depth images together
            depth_ims = None
            for i in range(num_images):
                start = time.time()
                small_color_im, new_depth_im, _ = sensor.frames()
                end = time.time()
                logging.info('Frames Runtime: %.3f' %(end-start))
                if depth_ims is None:
                    depth_ims = np.zeros([new_depth_im.height,
                                          new_depth_im.width,
                                          num_images])
                depth_ims[:,:,i] = new_depth_im.data

            med_depth_im = np.median(depth_ims, axis=2)
            depth_im = DepthImage(med_depth_im, sensor.ir_frame)

            # find the corner pixels in an upsampled version of the color image
            big_color_im = small_color_im.resize(color_image_rescale_factor)
            corner_px = big_color_im.find_chessboard(sx=sx, sy=sy)

            if vis:
                plt.figure()
                plt.imshow(big_color_im.data)
                for i in range(sx):
                    plt.scatter(corner_px[i,0], corner_px[i,1], s=25, c='b')
                plt.show()

            if corner_px is None:
                logging.error('No chessboard detected! Check camera exposure settings')
                continue

            # convert back to original image
            small_corner_px = corner_px / color_image_rescale_factor
        
            if vis:
                plt.figure()
                plt.imshow(small_color_im.data)
                for i in range(sx):
                    plt.scatter(small_corner_px[i,0], small_corner_px[i,1], s=25, c='b')
                plt.axis('off')
                plt.show()

            # project points into 3D
            camera_intr = sensor.ir_intrinsics
            points_3d = camera_intr.deproject(depth_im)

            # get round chessboard ind
            corner_px_round = np.round(small_corner_px).astype(np.uint16)
            corner_ind = depth_im.ij_to_linear(corner_px_round[:,0], corner_px_round[:,1])
            if corner_ind.shape[0] != sx*sy:
                logging.warning('Did not find all corners. Discarding...')
                continue

            # average 3d points
            points_3d_plane = (k * points_3d_plane + points_3d[corner_ind]) / (k + 1)
            logging.info('Registration iteration %d of %d' %(k+1, config['num_transform_avg']))
            k += 1

        # fit a plane to the chessboard corners
        X = np.c_[points_3d_plane.x_coords, points_3d_plane.y_coords, np.ones(points_3d_plane.num_points)]
        y = points_3d_plane.z_coords
        A = X.T.dot(X)
        b = X.T.dot(y)
        w = np.linalg.inv(A).dot(b)
        n = np.array([w[0], w[1], -1])
        n = n / np.linalg.norm(n)
        if flip_normal:
            n = -n
        mean_point_plane = points_3d_plane.mean()
        
        # find x-axis of the chessboard coordinates on the fitted plane
        T_camera_table = RigidTransform(translation = -points_3d_plane.mean().data,
                                    from_frame=points_3d_plane.frame,
                                    to_frame='table')
        points_3d_centered = T_camera_table * points_3d_plane

        # get points along y
        if point_order == 'row_major':
            coord_pos_x = int(math.floor(sx*sy/2.0))
            coord_neg_x = int(math.ceil(sx*sy/2.0))
            
            points_pos_x = points_3d_centered[coord_pos_x:]
            points_neg_x = points_3d_centered[:coord_neg_x]
            x_axis = np.mean(points_pos_x.data, axis=1) - np.mean(points_neg_x.data, axis=1)
            x_axis = x_axis - np.vdot(x_axis, n)*n
            x_axis = x_axis / np.linalg.norm(x_axis)
            y_axis = np.cross(n, x_axis)
        else:
            coord_pos_y = int(math.floor(sx*(sy-1)/2.0))
            coord_neg_y = int(math.ceil(sx*(sy+1)/2.0))
            points_pos_y = points_3d_centered[:coord_pos_y]
            points_neg_y = points_3d_centered[coord_neg_y:]
            y_axis = np.mean(points_pos_y.data, axis=1) - np.mean(points_neg_y.data, axis=1)
            y_axis = y_axis - np.vdot(y_axis, n)*n
            y_axis = y_axis / np.linalg.norm(y_axis)
            x_axis = np.cross(-n, y_axis)

        # produce translation and rotation from plane center and chessboard basis
        rotation_cb_camera = RigidTransform.rotation_from_axes(x_axis, y_axis, n)
        translation_cb_camera = mean_point_plane.data
        T_cb_camera = RigidTransform(rotation=rotation_cb_camera,
                                     translation=translation_cb_camera,
                                     from_frame='cb',
                                     to_frame=sensor.frame)
        
        if y_points_left and np.abs(T_cb_camera.y_axis[1]) > 0.1:
            if T_cb_camera.x_axis[0] > 0:
                T_cb_camera.rotation = T_cb_camera.rotation.dot(RigidTransform.z_axis_rotation(-np.pi/2).T)
            else:
                T_cb_camera.rotation = T_cb_camera.rotation.dot(RigidTransform.z_axis_rotation(np.pi/2).T)
        T_camera_cb = T_cb_camera.inverse()
                
        # optionally display cb corners with detected pose in 3d space
        if config['debug']:
            # display image with axes overlayed
            cb_center_im = camera_intr.project(Point(T_cb_camera.translation, frame=sensor.ir_frame))
            cb_x_im = camera_intr.project(Point(T_cb_camera.translation + T_cb_camera.x_axis * config['scale_amt'], frame=sensor.ir_frame))
            cb_y_im = camera_intr.project(Point(T_cb_camera.translation + T_cb_camera.y_axis * config['scale_amt'], frame=sensor.ir_frame))
            cb_z_im = camera_intr.project(Point(T_cb_camera.translation + T_cb_camera.z_axis * config['scale_amt'], frame=sensor.ir_frame))
            x_line = np.array([cb_center_im.data, cb_x_im.data])
            y_line = np.array([cb_center_im.data, cb_y_im.data])
            z_line = np.array([cb_center_im.data, cb_z_im.data])

            plt.figure()
            plt.imshow(small_color_im.data)
            plt.scatter(cb_center_im.data[0], cb_center_im.data[1])
            plt.plot(x_line[:,0], x_line[:,1], c='r', linewidth=3)
            plt.plot(y_line[:,0], y_line[:,1], c='g', linewidth=3)
            plt.plot(z_line[:,0], z_line[:,1], c='b', linewidth=3)
            plt.axis('off')
            plt.title('Chessboard frame in camera %s' %(sensor.frame))
            plt.show()

        return ChessboardRegistrationResult(T_camera_cb, points_3d_plane)



"""
Abstract class for Camera sensors.
Author: Jeff Mahler
"""
from abc import ABCMeta, abstractmethod

from .camera_intrinsics import CameraIntrinsics
from .constants import *
from .image import ColorImage, DepthImage, IrImage

import os

class CameraSensor(object):
    """Abstract base class for camera sensors.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def start(self):
        """Starts the sensor stream.
        """
        pass

    @abstractmethod
    def stop(self):
        """Stops the sensor stream.
        """
        pass

    def reset(self):
        """Restarts the sensor stream.
        """
        self.stop()
        self.start()

    @abstractmethod
    def frames(self):
        """Returns the latest set of frames.
        """
        pass


class VirtualSensor(CameraSensor):
    SUPPORTED_FILE_EXTS = ['.png', '.npy']

    """ Class for a virtual sensor that uses pre-captured images
    stored to disk instead of actually connecting to a sensor.
    For debugging purposes.
    """
    def __init__(self, path_to_images, frame=None, loop=True):
        """Create a new virtualized Primesense sensor.

        This requires a directory containing a specific set of files.

        First, the directory must contain a set of images, where each
        image has three files:
        - color_{#}.png
        - depth_{#}.npy
        - ir_{#}.npy
        In these, the {#} is replaced with the integer index for the
        image. These indices should start at zero and increase
        consecutively.

        Second, the directory must contain CameraIntrisnics files
        for the color and ir cameras:
        - {frame}_color.intr
        - {frame}_ir.intr
        In these, the {frame} is replaced with the reference frame
        name that is passed as a parameter to this function.

        Parameters
        ----------
        path_to_images : :obj:`str`
            The path to a directory containing images that the virtualized
            sensor will return.
        frame : :obj:`str`
            The name of the frame of reference in which the sensor resides.
            If None, this will be discovered from the files in the directory.
        loop : :obj:`str`
            Whether or not to loop back to the first image after running out
        """
        self._running = False
        self._path_to_images = path_to_images
        self._im_index = 0
        self._num_images = 0
        self._frame = frame
        filenames = os.listdir(self._path_to_images)

        # get number of images
        for filename in filenames:
            if filename.find('depth') != -1 and filename.endswith('.npy'):
                self._num_images += 1

        # set the frame dynamically
        self._color_ext = '.png'
        for filename in filenames:
            file_root, file_ext = os.path.splitext(filename)
            color_ind = file_root.rfind('color')

            if file_ext in VirtualSensor.SUPPORTED_FILE_EXTS \
               and color_ind != -1:
                self._color_ext = file_ext

            if self._frame is None and file_ext == INTR_EXTENSION and color_ind != -1:
                self._frame = file_root[:color_ind-1]
                self._color_frame = file_root
                self._ir_frame = file_root

        # load color intrinsics
        color_intr_filename = os.path.join(self._path_to_images, '%s_color.intr' %(self._frame))
        ir_intr_filename = os.path.join(self._path_to_images, '%s_ir.intr' %(self._frame))
        generic_intr_filename = os.path.join(self._path_to_images, '%s.intr' %(self._frame))
        if os.path.exists(color_intr_filename):
            self._color_intr = CameraIntrinsics.load(color_intr_filename)
        else:
            self._color_intr = CameraIntrinsics.load(generic_intr_filename)
        if os.path.exists(ir_intr_filename):            
            self._ir_intr = CameraIntrinsics.load(ir_intr_filename)
        else:
            self._ir_intr = CameraIntrinsics.load(generic_intr_filename)            

    @property
    def path_to_images(self):
        """:obj:`str` : The path to a directory containing images that the virtualized
        sensor will return.
        """
        return self._path_to_images

    @property
    def is_running(self):
        """bool : True if the stream is running, or false otherwise.
        """
        return self._running

    @property
    def frame(self):
        """:obj:`str` : The reference frame of the sensor.
        """
        return self._frame

    @property
    def color_frame(self):
        """:obj:`str` : The reference frame of the color sensor.
        """
        return self._color_frame

    @property
    def ir_frame(self):
        """:obj:`str` : The reference frame of the IR sensor.
        """
        return self._ir_frame

    @property
    def color_intrinsics(self):
        """:obj:`CameraIntrinsics` : The camera intrinsics for the sensor's color camera.
        """
        return self._color_intr

    @property
    def ir_intrinsics(self):
        """:obj:`CameraIntrinsics` : The camera intrinsics for the sensor's IR camera.
        """
        return self._ir_intr


    def start(self):
        """Starts the sensor stream.

        In this virtualized sensor, this simply resets the image index to zero.
        Everytime start is called, we start the stream again at the first image.
        """
        self._im_index = 0
        self._running = True

    def stop(self):
        """Stops the sensor stream.

        Returns
        -------
        bool
            True if the stream was stopped, False if the device was already
            stopped or was not otherwise available.
        """
        if not self._running:
            return False
        self._running = False
        return True

    def frames(self):
        """Retrieve the next frame from the image directory and convert it to a ColorImage,
        a DepthImage, and an IrImage.

        Parameters
        ----------
        skip_registration : bool
            If True, the registration step is skipped.

        Returns
        -------
        :obj:`tuple` of :obj:`ColorImage`, :obj:`DepthImage`, :obj:`IrImage`, :obj:`numpy.ndarray`
            The ColorImage, DepthImage, and IrImage of the current frame.

        Raises
        ------
        RuntimeError
            If the stream is not running or if all images in the
            directory have been used.
        """
        if not self._running:
            raise RuntimeError('Device pointing to %s not runnning. Cannot read frames' %(self._path_to_images))

        if self._im_index >= self._num_images:
            raise RuntimeError('Device is out of images')

        # read images
        color_filename = os.path.join(self._path_to_images, 'color_%d%s' %(self._im_index, self._color_ext))
        color_im = ColorImage.open(color_filename, frame=self._frame)
        depth_filename = os.path.join(self._path_to_images, 'depth_%d.npy' %(self._im_index))
        depth_im = DepthImage.open(depth_filename, frame=self._frame)
        self._im_index = (self._im_index + 1) % self._num_images
        return color_im, depth_im, None

    def median_depth_img(self, num_img=1):
        """Collect a series of depth images and return the median of the set.

        Parameters
        ----------
        num_img : int
            The number of consecutive frames to process.

        Returns
        -------
        :obj:`DepthImage`
            The median DepthImage collected from the frames.
        """
        depths = []

        for _ in range(num_img):
            _, depth, _ = self.frames()
            depths.append(depth)

        return Image.median_images(depths)
    

class TensorDatasetVirtualSensor(VirtualSensor):
    CAMERA_INTR_FIELD = 'camera_intrs'
    COLOR_IM_FIELD = 'color_ims'
    DEPTH_IM_FIELD = 'depth_ims'
    IMAGE_FIELDS = [COLOR_IM_FIELD, DEPTH_IM_FIELD]
    
    """ Class for a virtual sensor that runs off of images stored in a
    tensor dataset.
    """
    def __init__(self, dataset_path, frame=None, loop=True):
        self._dataset_path = dataset_path
        self._frame = frame
        self._color_frame = frame
        self._ir_frame = frame
        self._im_index = 0
        self._running = False
        
        from dexnet.learning import TensorDataset
        self._dataset = TensorDataset.open(self._dataset_path)
        self._num_images = self._dataset.num_datapoints
        self._image_rescale_factor = 1.0
        if 'image_rescale_factor' in self._dataset.metadata.keys():
            self._image_rescale_factor = 1.0 / self._dataset.metadata['image_rescale_factor']
        
        datapoint = self._dataset.datapoint(0, [TensorDatasetVirtualSensor.CAMERA_INTR_FIELD])
        camera_intr_vec = datapoint[TensorDatasetVirtualSensor.CAMERA_INTR_FIELD]
        self._color_intr = CameraIntrinsics.from_vec(camera_intr_vec, frame=self._color_frame).resize(self._image_rescale_factor)
        self._ir_intr = CameraIntrinsics.from_vec(camera_intr_vec, frame=self._ir_frame).resize(self._image_rescale_factor)

    def frames(self):
        """Retrieve the next frame from the tensor dataset and convert it to a ColorImage,
        a DepthImage, and an IrImage.

        Parameters
        ----------
        skip_registration : bool
            If True, the registration step is skipped.

        Returns
        -------
        :obj:`tuple` of :obj:`ColorImage`, :obj:`DepthImage`, :obj:`IrImage`, :obj:`numpy.ndarray`
            The ColorImage, DepthImage, and IrImage of the current frame.

        Raises
        ------
        RuntimeError
            If the stream is not running or if all images in the
            directory have been used.
        """
        if not self._running:
            raise RuntimeError('Device pointing to %s not runnning. Cannot read frames' %(self._path_to_images))

        if self._im_index >= self._num_images:
            raise RuntimeError('Device is out of images')

        # read images
        datapoint = self._dataset.datapoint(self._im_index,
                                            TensorDatasetVirtualSensor.IMAGE_FIELDS)
        color_im = ColorImage(datapoint[TensorDatasetVirtualSensor.COLOR_IM_FIELD],
                              frame=self._frame)
        depth_im = DepthImage(datapoint[TensorDatasetVirtualSensor.DEPTH_IM_FIELD],
                              frame=self._frame)
        if self._image_rescale_factor != 1.0:
            color_im = color_im.resize(self._image_rescale_factor)
            depth_im = depth_im.resize(self._image_rescale_factor, interp='nearest')
        self._im_index = (self._im_index + 1) % self._num_images
        return color_im, depth_im, None



"""
Wrapper for rendered images of an object on a table using Maya or our stable pose
Author: Jeff Mahler
"""
from autolab_core import RigidTransform

from .image import BinaryImage, ColorImage, DepthImage

class RenderMode(object):
    """Supported rendering modes.
    """
    SEGMASK = 'segmask'
    DEPTH = 'depth'
    SCALED_DEPTH = 'scaled_depth'
    COLOR = 'color'
    GRAY = 'gray'
    GD = 'gd'
    RGBD = 'rgbd'
    GD = 'gd'
    GRAYSCALE = 'gray'

class ObjectRender(object):
    """Class to encapsulate images of an object rendered from a virtual camera.

    Note
    ----
        In this class, the table's frame of reference is the 'world' frame for
        the renderer.
    """

    def __init__(self, image, T_camera_world=RigidTransform(from_frame='camera', to_frame='table'),
                 obj_key = None, stable_pose=None):
        """Create an ObjectRender.

        Parameters
        ----------
        image : :obj:`Image`
            The image to be encapsulated.

        T_camera_world : :obj:`autolab_core.RigidTransform`
            A rigid transform from camera to world coordinates (positions the
            camera in the world). TODO -- this should be renamed.

        obj_key : :obj:`str`, optional
            A string identifier for the object being rendered.

        stable_pose : :obj:`meshpy.StablePose`
            The object's stable pose.
        """
        self.image = image
        self.T_camera_world = T_camera_world
        self.obj_key = obj_key
        self.stable_pose = stable_pose

    @property
    def T_obj_camera(self):
        """Returns the transformation from camera to object when the object is in the given stable pose.

        Returns
        -------
        :obj:`autolab_core.RigidTransform`
            The desired transform.
        """
        if self.stable_pose is None:
            T_obj_world = RigidTransform(from_frame='obj', to_frame='world')
        else:
            T_obj_world = self.stable_pose.T_obj_table.as_frames('obj', 'world')
        T_camera_obj = T_obj_world.inverse() * self.T_camera_world
        return T_camera_obj

class QueryImageBundle:
    """Struct to encapsulate renders of multiple image types.
    """
    def __init__(self, binary_image_render, color_image_render, depth_image_render):
        """Initialize a QueryImageBundle.

        Parameters
        ----------
        binary_image_render : :obj:`ObjectRender`
            A render of a :obj:`BinaryImage`.

        color_image_render : :obj:`ObjectRender`
            A render of a :obj:`ColorImage`.

        depth_image_render : :obj:`ObjectRender`
            A render of a :obj:`DepthImage`.
        """
        self.binary_im_render = binary_image_render
        self.color_im_render = color_image_render
        self.depth_im_render = depth_image_render

    @property
    def binary_im(self):
        """:obj:`BinaryImage` : The binary image from the set.
        """
        return self.binary_im_render.image

    @property
    def color_im(self):
        """:obj:`ColorImage` : The color image from the set.
        """
        return self.color_im_render.image

    @property
    def depth_im(self):
        """:obj:`DepthImage` : The depth image from the set.
        """
        return self.depth_im_render.image

    def image(self, render_mode):
        """Return an image generated with a particular render mode.

        Parameters
        ----------
        render_mode : :obj:`RenderMode`
            The type of image we want.

        Returns
        -------
        :obj:`Image`
            The color, depth, or binary image if render_mode is
            COLOR, DEPTH, or SEGMASK respectively.
        """
        if render_mode == RenderMode.COLOR:
            return self.color_im
        elif render_mode == RenderMode.DEPTH:
            return self.depth_im
        elif render_mode == RenderMode.SEGMASK:
            return self.binary_im
        else:
            return None



"""
Classes for image feature extraction.
Author: Jeff Mahler
"""
from abc import ABCMeta, abstractmethod

import logging
import numpy as np
import time

from .constants import *
from .cnn import AlexNet
from .image import Image, ColorImage, DepthImage

class FeatureExtractor:
    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    @abstractmethod
    def extract(self, image):
        """
        Returns a set of extracted features for an image
        Parameters
        ----------
        image : :obj:`Image`
            object to extract for
        """
        pass

class CNNBatchFeatureExtractor(FeatureExtractor):
    """ Extract feature descriptors for images in a giant batch using Convolutional Neural Networks.

    Attributes
    ----------
    cnn : :obj:`AlexNet`
        the convolutional neural network to use
    """
    def __init__(self, config):
        self.cnn_ = AlexNet(config, use_default_weights=True)

    def open(self):
        """ Opens the tensorflow session. For memory management. """
        self.cnn_.open_session()

    def close(self):
        """ Closes the tensorflow session. For memory management. """
        self.cnn_.close_session()

    def _forward_pass(self, images):
        """ Forward pass a list of images through the CNN """
        # form image array
        num_images = len(images)
        if num_images == 0:
            return None
        for image in images:
            if not isinstance(image, Image):
                new_images = []
                for image in images:
                    if len(image.shape) > 2:
                        new_images.append(ColorImage(image, frame='unspecified'))
                    elif image.dtype == np.float32 or image.dtype == np.float64:
                        new_images.append(DepthImage(image, frame='unspecified'))
                    else:
                        raise ValueError('Image type not understood')
                images = new_images
                break

        im_height = images[0].height
        im_width = images[0].width
        channels = images[0].channels
        tensor_channels = 3
        image_arr = np.zeros([num_images, im_height, im_width, tensor_channels])
        for j, image in enumerate(images):
            if channels == 3:
                image_arr[j,:,:,:] = image.raw_data
            else:
                image_arr[j,:,:,:] = np.tile(image.raw_data, [1,1,1,3])

        # predict
        fp_start = time.time()
        final_blobs = self.cnn_.featurize(image_arr)
        fp_stop = time.time()
        logging.debug('Featurization took %f sec per image' %((fp_stop - fp_start) / len(images)))
        return final_blobs.reshape(final_blobs.shape[0], -1)

    def extract(self, images):
        """ Form feature descriptors for a set of images.

        Parameters
        ----------
        images : :obj:`list` of :obj`Image` or :obj`numpy.ndarray`
            images to extract features for
        """
        return self._forward_pass(images)

class CNNReusableBatchFeatureExtractor(CNNBatchFeatureExtractor):
    """ Extract feature descriptors for images in a giant batch. Allows you to initialize the extractor with a pre-existing CNN, for memory management reasons. """
    def __init__(self, cnn):
        self.cnn_ = cnn



"""
Class for interfacing with the Primesense Carmine RGBD sensor
Author: Jeff Mahler
"""
import copy
import logging
import numpy as np
import os

from .constants import MM_TO_METERS, INTR_EXTENSION
try:
    from primesense import openni2
except:
    logging.warning('Unable to import openni2 driver. Python-only Primesense driver may not work properly')

from perception import CameraIntrinsics, CameraSensor, ColorImage, DepthImage, IrImage, Image

class PrimesenseRegistrationMode:
    """Primesense registration mode.
    """
    NONE = 0
    DEPTH_TO_COLOR = 1

class PrimesenseSensor(CameraSensor):
    """ Class for interacting with a Primesense RGBD sensor.
    """
    #Constants for image height and width (in case they're needed somewhere)
    COLOR_IM_HEIGHT = 480
    COLOR_IM_WIDTH = 640
    DEPTH_IM_HEIGHT = 480
    DEPTH_IM_WIDTH = 640
    CENTER_X = float(DEPTH_IM_WIDTH-1) / 2.0
    CENTER_Y = float(DEPTH_IM_HEIGHT-1) / 2.0
    FOCAL_X = 525.
    FOCAL_Y = 525.
    FPS = 30
    OPENNI2_PATH = '/home/autolab/Libraries/OpenNI-Linux-x64-2.2/Redist'

    def __init__(self, registration_mode=PrimesenseRegistrationMode.DEPTH_TO_COLOR,
                 auto_white_balance=False, auto_exposure=True,
                 enable_depth_color_sync=True, flip_images=True, frame=None):
        self._device = None
        self._depth_stream = None
        self._color_stream = None
        self._running = None

        self._registration_mode = registration_mode
        self._auto_white_balance = auto_white_balance
        self._auto_exposure = auto_exposure
        self._enable_depth_color_sync = enable_depth_color_sync
        self._flip_images = flip_images

        self._frame = frame

        if self._frame is None:
            self._frame = 'primesense'
        self._color_frame = '%s_color' %(self._frame)
        self._ir_frame = self._frame # same as color since we normally use this one

    def __del__(self):
        """Automatically stop the sensor for safety.
        """
        if self.is_running:
            self.stop()

    @property
    def color_intrinsics(self):
        """:obj:`CameraIntrinsics` : The camera intrinsics for the primesense color camera.
        """
        return CameraIntrinsics(self._ir_frame, PrimesenseSensor.FOCAL_X, PrimesenseSensor.FOCAL_Y,
                                PrimesenseSensor.CENTER_X, PrimesenseSensor.CENTER_Y,
                                height=PrimesenseSensor.DEPTH_IM_HEIGHT,
                                width=PrimesenseSensor.DEPTH_IM_WIDTH)

    @property
    def ir_intrinsics(self):
        """:obj:`CameraIntrinsics` : The camera intrinsics for the primesense IR camera.
        """
        return CameraIntrinsics(self._ir_frame, PrimesenseSensor.FOCAL_X, PrimesenseSensor.FOCAL_Y,
                                PrimesenseSensor.CENTER_X, PrimesenseSensor.CENTER_Y,
                                height=PrimesenseSensor.DEPTH_IM_HEIGHT,
                                width=PrimesenseSensor.DEPTH_IM_WIDTH)

    @property
    def is_running(self):
        """bool : True if the stream is running, or false otherwise.
        """
        return self._running

    @property
    def frame(self):
        """:obj:`str` : The reference frame of the sensor.
        """
        return self._frame

    @property
    def color_frame(self):
        """:obj:`str` : The reference frame of the color sensor.
        """
        return self._color_frame

    @property
    def ir_frame(self):
        """:obj:`str` : The reference frame of the IR sensor.
        """
        return self._ir_frame
    
    def start(self):
        """ Start the sensor """
        # open device
        openni2.initialize(PrimesenseSensor.OPENNI2_PATH)
        self._device = openni2.Device.open_any()

        # open depth stream
        self._depth_stream = self._device.create_depth_stream()
        self._depth_stream.configure_mode(PrimesenseSensor.DEPTH_IM_WIDTH,
                                          PrimesenseSensor.DEPTH_IM_HEIGHT,
                                          PrimesenseSensor.FPS,
                                          openni2.PIXEL_FORMAT_DEPTH_1_MM) 
        self._depth_stream.start()

        # open color stream
        self._color_stream = self._device.create_color_stream()
        self._color_stream.configure_mode(PrimesenseSensor.COLOR_IM_WIDTH,
                                          PrimesenseSensor.COLOR_IM_HEIGHT,
                                          PrimesenseSensor.FPS,
                                          openni2.PIXEL_FORMAT_RGB888) 
        self._color_stream.camera.set_auto_white_balance(self._auto_white_balance)
        self._color_stream.camera.set_auto_exposure(self._auto_exposure)
        self._color_stream.start()

        # configure device
        if self._registration_mode == PrimesenseRegistrationMode.DEPTH_TO_COLOR:
            self._device.set_image_registration_mode(openni2.IMAGE_REGISTRATION_DEPTH_TO_COLOR)
        else:
            self._device.set_image_registration_mode(openni2.IMAGE_REGISTRATION_OFF)

        self._device.set_depth_color_sync_enabled(self._enable_depth_color_sync)

        self._running = True

    def stop(self):
        """ Stop the sensor """
        # check that everything is running
        if not self._running or self._device is None:
            logging.warning('Primesense not running. Aborting stop')
            return False

        # stop streams
        if self._depth_stream:
            self._depth_stream.stop()
        if self._color_stream:
            self._color_stream.stop()
        self._running = False
        
        # Unload openni2
        openni2.unload()
        return True

    def _read_depth_image(self):
        """ Reads a depth image from the device """
        # read raw uint16 buffer
        im_arr = self._depth_stream.read_frame()
        raw_buf = im_arr.get_buffer_as_uint16()
        buf_array = np.array([raw_buf[i] for i in range(PrimesenseSensor.DEPTH_IM_WIDTH * PrimesenseSensor.DEPTH_IM_HEIGHT)])

        # convert to image in meters
        depth_image = buf_array.reshape(PrimesenseSensor.DEPTH_IM_HEIGHT,
                                        PrimesenseSensor.DEPTH_IM_WIDTH)
        depth_image = depth_image * MM_TO_METERS # convert to meters
        if self._flip_images:
            depth_image = np.flipud(depth_image)
        else:
            depth_image = np.fliplr(depth_image)
        return DepthImage(depth_image, frame=self._frame)

    def _read_color_image(self):
        """ Reads a color image from the device """
        # read raw buffer
        im_arr = self._color_stream.read_frame()
        raw_buf = im_arr.get_buffer_as_triplet()
        r_array = np.array([raw_buf[i][0] for i in range(PrimesenseSensor.COLOR_IM_WIDTH * PrimesenseSensor.COLOR_IM_HEIGHT)])        
        g_array = np.array([raw_buf[i][1] for i in range(PrimesenseSensor.COLOR_IM_WIDTH * PrimesenseSensor.COLOR_IM_HEIGHT)])        
        b_array = np.array([raw_buf[i][2] for i in range(PrimesenseSensor.COLOR_IM_WIDTH * PrimesenseSensor.COLOR_IM_HEIGHT)])        

        # convert to uint8 image
        color_image = np.zeros([PrimesenseSensor.COLOR_IM_HEIGHT, PrimesenseSensor.COLOR_IM_WIDTH, 3])
        color_image[:,:,0] = r_array.reshape(PrimesenseSensor.COLOR_IM_HEIGHT,
                                             PrimesenseSensor.COLOR_IM_WIDTH)
        color_image[:,:,1] = g_array.reshape(PrimesenseSensor.COLOR_IM_HEIGHT,
                                             PrimesenseSensor.COLOR_IM_WIDTH)
        color_image[:,:,2] = b_array.reshape(PrimesenseSensor.COLOR_IM_HEIGHT,
                                             PrimesenseSensor.COLOR_IM_WIDTH)
        if self._flip_images:
            color_image = np.flipud(color_image.astype(np.uint8))
        else:
            color_image = np.fliplr(color_image.astype(np.uint8))
        return ColorImage(color_image, frame=self._frame)

    def frames(self):
        """Retrieve a new frame from the Kinect and convert it to a ColorImage,
        a DepthImage, and an IrImage.

        Returns
        -------
        :obj:`tuple` of :obj:`ColorImage`, :obj:`DepthImage`, :obj:`IrImage`, :obj:`numpy.ndarray`
            The ColorImage, DepthImage, and IrImage of the current frame.

        Raises
        ------
        RuntimeError
            If the Kinect stream is not running.
        """
        color_im = self._read_color_image()
        depth_im = self._read_depth_image()
        return color_im, depth_im, None

    def median_depth_img(self, num_img=1, fill_depth=0.0):
        """Collect a series of depth images and return the median of the set.

        Parameters
        ----------
        num_img : int
            The number of consecutive frames to process.

        Returns
        -------
        :obj:`DepthImage`
            The median DepthImage collected from the frames.
        """
        depths = []

        for _ in range(num_img):
            _, depth, _ = self.frames()
            depths.append(depth)

        median_depth = Image.median_images(depths)
        median_depth.data[median_depth.data == 0.0] = fill_depth
        return median_depth

    def min_depth_img(self, num_img=1):
        """Collect a series of depth images and return the min of the set.

        Parameters
        ----------
        num_img : int
            The number of consecutive frames to process.

        Returns
        -------
        :obj:`DepthImage`
            The min DepthImage collected from the frames.
        """
        depths = []

        for _ in range(num_img):
            _, depth, _ = self.frames()
            depths.append(depth)

        return Image.min_images(depths)

class PrimesenseSensor_ROS(PrimesenseSensor):
    """ ROS-based version of Primesense RGBD sensor interface
    
    Requires starting the openni2 ROS driver and the two stream_image_buffer (image_buffer.py)
    ros services for depth and color images. By default, the class will look for the depth_image buffer
    and color_image buffers under "{frame}/depth/stream_image_buffer" and "{frame}/rgb/stream_image_buffer"
    respectively (within the current ROS namespace).
    
    This can be changed by passing in depth_image_buffer, color_image_buffer (which change where the program
    looks for the buffer services) and depth_absolute, color_absolute (which changes whether the program prepends
    the current ROS namespace).
    """
    def __init__(self, depth_image_buffer= None, depth_absolute=False, color_image_buffer=None, color_absolute=False,
                 flip_images=True, frame=None, staleness_limit=10., timeout=10):  
        import rospy
        from rospy import numpy_msg
        from perception.srv import *
        ImageBufferResponse = rospy.numpy_msg.numpy_msg(ImageBufferResponse)
        ImageBuffer._response_class = ImageBufferResponse
    
        self._flip_images = flip_images
        self._frame = frame
        
        self.staleness_limit = staleness_limit
        self.timeout = timeout

        if self._frame is None:
            self._frame = 'primesense'
        self._color_frame = '%s_color' %(self._frame)
        self._ir_frame = self._frame # same as color since we normally use this one
        
        # Set image buffer locations
        self._depth_image_buffer = ('{0}/depth/stream_image_buffer'.format(frame)
                                    if depth_image_buffer == None else depth_image_buffer)
        self._color_image_buffer = ('{0}/rgb/stream_image_buffer'.format(frame)
                                    if color_image_buffer == None else color_image_buffer)
        if not depth_absolute:
            self._depth_image_buffer = rospy.get_namespace() + self._depth_image_buffer
        if not color_absolute:
            self._color_image_buffer = rospy.get_namespace() + self._color_image_buffer
        
    def start(self):
        """For PrimesenseSensor, start/stop by launching/stopping
        the associated ROS services"""
        pass
    def stop(self):
        """For PrimesenseSensor, start/stop by launching/stopping
        the associated ROS services"""
        pass
    
    def _ros_read_images(self, stream_buffer, number, staleness_limit = 10.):
        """ Reads images from a stream buffer
        
        Parameters
        ----------
        stream_buffer : string
            absolute path to the image buffer service
        number : int
            The number of frames to get. Must be less than the image buffer service's
            current buffer size
        staleness_limit : float, optional
            Max value of how many seconds old the oldest image is. If the oldest image
            grabbed is older than this value, a RuntimeError is thrown.
            
            If None, staleness is ignored.
        Returns
        -------
        List of nump.ndarray objects, each one an image
        Images are in reverse chronological order (newest first)
        """
        
        rospy.wait_for_service(stream_buffer, timeout = self.timeout)
        ros_image_buffer = rospy.ServiceProxy(stream_buffer, ImageBuffer)
        ret = ros_image_buffer(number, 1)
        if not staleness_limit == None:
            if ret.timestamps[-1] > staleness_limit:
                raise RuntimeError("Got data {0} seconds old, more than allowed {1} seconds"
                                   .format(ret.timestamps[-1], staleness_limit))
            
        data = ret.data.reshape(ret.data_dim1, ret.data_dim2, ret.data_dim3).astype(ret.dtype)
        
        # Special handling for 1 element, since dstack's behavior is different
        if number == 1:
            return [data]
        return np.dsplit(data, number)

    @property
    def is_running(self):
        """bool : True if the image buffers are running, or false otherwise.
        
        Does this by grabbing one frame with staleness checking
        """
        try:
            self.frames()
        except:
            return False
        return True
    
    def _read_depth_images(self, num_images):
        """ Reads depth images from the device """
        depth_images = self._ros_read_images(self._depth_image_buffer, num_images, self.staleness_limit)
        for i in range(0, num_images):
            depth_images[i] = depth_images[i] * MM_TO_METERS # convert to meters
            if self._flip_images:
                depth_images[i] = np.flipud(depth_images[i])
                depth_images[i] = np.fliplr(depth_images[i])
            depth_images[i] = DepthImage(depth_images[i], frame=self._frame) 
        return depth_images
    def _read_color_images(self, num_images):
        """ Reads color images from the device """
        color_images = self._ros_read_images(self._color_image_buffer, num_images, self.staleness_limit)
        for i in range(0, num_images):
            if self._flip_images:
                color_images[i] = np.flipud(color_images[i].astype(np.uint8))
                color_images[i] = np.fliplr(color_images[i].astype(np.uint8))
            color_images[i] = ColorImage(color_images[i], frame=self._frame) 
        return color_images
    
    def _read_depth_image(self):
        """ Wrapper to maintain compatibility """
        return self._read_depth_images(1)[0]
    def _read_color_image(self):
        """ Wrapper to maintain compatibility """
        return self._read_color_images(1)[0]
        
    def median_depth_img(self, num_img=1, fill_depth=0.0):
        """Collect a series of depth images and return the median of the set.

        Parameters
        ----------
        num_img : int
            The number of consecutive frames to process.

        Returns
        -------
        :obj:`DepthImage`
            The median DepthImage collected from the frames.
        """
        depths = self._read_depth_images(num_img)

        median_depth = Image.median_images(depths)
        median_depth.data[median_depth.data == 0.0] = fill_depth
        return median_depth

    def min_depth_img(self, num_img=1):
        """Collect a series of depth images and return the min of the set.

        Parameters
        ----------
        num_img : int
            The number of consecutive frames to process.

        Returns
        -------
        :obj:`DepthImage`
            The min DepthImage collected from the frames.
        """
        depths = self._read_depth_images(num_img)

        return Image.min_images(depths)
        



"""
Global constants for the Perception module.
"""
METERS_TO_MM = 1000.0
MM_TO_METERS = 1.0 / METERS_TO_MM

COLOR_IMAGE_EXTS = ['.png', '.jpg']
MIN_DEPTH = 0.25
MAX_DEPTH = 1.25
MAX_IR = 65535

TF_EXTENSION = '.tf'
INTR_EXTENSION = '.intr'



import logging
import numpy as np
import os
import time

from autolab_core import RigidTransform, PointCloud
from . import CameraSensor, ColorImage, CameraIntrinsics, PhoXiSensor, WebcamSensor

class ColorizedPhoXiSensor(CameraSensor):
    """Class for using a Logitech Webcam sensor to colorize a PhoXi's point clouds.
    """

    def __init__(self, phoxi_config, webcam_config, calib_dir, frame='phoxi'):
        """Initialize a webcam-colorized PhoXi sensor.

        Parameters
        ----------
        frame : str
            A name for the frame in which images are returned.
        phoxi_config : dict
            Config for the PhoXi camera.
        phoxi_to_world_fn : str
            Filepath for T_phoxi_world.
        webcam_config : dict
            Config for the webcam.
        webcam_to_world_fn : str
            Filepath for T_webcam_world
        """
        self._frame = frame
        phoxi_to_world_fn = os.path.join(calib_dir, 'phoxi', 'phoxi_to_world.tf')
        webcam_to_world_fn = os.path.join(calib_dir, 'webcam', 'webcam_to_world.tf')
        self._T_phoxi_world = RigidTransform.load(phoxi_to_world_fn)
        self._T_webcam_world = RigidTransform.load(webcam_to_world_fn)
        self._phoxi = PhoXiSensor(**phoxi_config)
        self._webcam = WebcamSensor(**webcam_config)
        self._camera_intr = self._phoxi.ir_intrinsics
        self._running = False

    def __del__(self):
        """Automatically stop the sensor for safety.
        """
        if self.is_running:
            self.stop()

    @property
    def color_intrinsics(self):
        """CameraIntrinsics : The camera intrinsics for the PhoXi Greyscale camera.
        """
        return self._camera_intr

    @property
    def ir_intrinsics(self):
        """CameraIntrinsics : The camera intrinsics for the PhoXi IR camera.
        """
        return self._camera_intr

    @property
    def is_running(self):
        """bool : True if the stream is running, or false otherwise.
        """
        return self._running

    @property
    def frame(self):
        """str : The reference frame of the sensor.
        """
        return self._frame

    @property
    def color_frame(self):
        """str : The reference frame of the sensor.
        """
        return self._frame

    @property
    def ir_frame(self):
        """str : The reference frame of the sensor.
        """
        return self._frame

    def start(self):
        """Start the sensor.
        """
        running = self._webcam.start()
        if not running:
            return running

        running &= self._phoxi.start()
        if not running:
            self._webcam.stop()
        return running

    def stop(self):
        """Stop the sensor.
        """
        # Check that everything is running
        if not self._running:
            logging.warning('Colorized PhoXi not running. Aborting stop')
            return False

        self._webcam.stop()
        self._phoxi.stop()

        return True

    def frames(self):
        """Retrieve a new frame from the PhoXi and convert it to a ColorImage,
        a DepthImage, and an IrImage.

        Returns
        -------
        :obj:`tuple` of :obj:`ColorImage`, :obj:`DepthImage`, :obj:`IrImage`, :obj:`numpy.ndarray`
            The ColorImage, DepthImage, and IrImage of the current frame.
        """
        _, phoxi_depth_im, _ = self._phoxi.frames()
        webcam_color_im, _, _ = self._webcam.frames(most_recent=True)

        # Colorize PhoXi Image
        phoxi_color_im = self._colorize(phoxi_depth_im, webcam_color_im)
        return phoxi_color_im, phoxi_depth_im, None

    def median_depth_img(self, num_img=1, fill_depth=0.0):
        """Collect a series of depth images and return the median of the set.

        Parameters
        ----------
        num_img : int
            The number of consecutive frames to process.

        Returns
        -------
        DepthImage
            The median DepthImage collected from the frames.
        """
        depths = []

        for _ in range(num_img):
            _, depth, _ = self.frames()
            depths.append(depth)

        median_depth = Image.median_images(depths)
        median_depth.data[median_depth.data == 0.0] = fill_depth
        return median_depth

    def _colorize(self, depth_im, color_im):
        """Colorize a depth image from the PhoXi using a color image from the webcam.

        Parameters
        ----------
        depth_im : DepthImage
            The PhoXi depth image.
        color_im : ColorImage
            Corresponding color image.

        Returns
        -------
        ColorImage
            A colorized image corresponding to the PhoXi depth image.
        """
        # Project the point cloud into the webcam's frame
        target_shape = (depth_im.data.shape[0], depth_im.data.shape[1], 3)
        pc_depth = self._phoxi.ir_intrinsics.deproject(depth_im)
        pc_color = self._T_webcam_world.inverse().dot(self._T_phoxi_world).apply(pc_depth)

        # Sort the points by their distance from the webcam's apeture
        pc_data = pc_color.data.T
        dists = np.linalg.norm(pc_data, axis=1)
        order = np.argsort(dists)
        pc_data = pc_data[order]
        pc_color = PointCloud(pc_data.T, frame=self._webcam.color_intrinsics.frame)
        sorted_dists = dists[order]
        sorted_depths = depth_im.data.flatten()[order]

        # Generate image coordinates for each sorted point
        icds = self._webcam.color_intrinsics.project(pc_color).data.T

        # Create mask for points that are masked by others
        rounded_icds = np.array(icds / 3.0, dtype=np.uint32)
        unique_icds, unique_inds, unique_inv = np.unique(rounded_icds, axis=0, return_index=True, return_inverse=True)
        icd_depths = sorted_dists[unique_inds]
        min_depths_pp = icd_depths[unique_inv]
        depth_delta_mask = np.abs(min_depths_pp - sorted_dists) < 5e-3

        # Create mask for points with missing depth or that lie outside the image
        valid_mask = np.logical_and(np.logical_and(icds[:,0] >= 0, icds[:,0] < self._webcam.color_intrinsics.width),
                                    np.logical_and(icds[:,1] >= 0, icds[:,1] < self._webcam.color_intrinsics.height))
        valid_mask = np.logical_and(valid_mask, sorted_depths != 0.0)
        valid_mask = np.logical_and(valid_mask, depth_delta_mask)
        valid_icds = icds[valid_mask]

        colors = color_im.data[valid_icds[:,1],valid_icds[:,0],:]
        color_im_data = np.zeros((target_shape[0] * target_shape[1], target_shape[2]), dtype=np.uint8)
        color_im_data[valid_mask] = colors
        color_im_data[order] = color_im_data.copy()
        color_im_data = color_im_data.reshape(target_shape)
        return ColorImage(color_im_data, frame=self._frame)





"""
Wrapper for tensorflow Convolutional Neural Network classes
Author: Jeff Mahler
"""
import numpy as np
import os
import tensorflow as tf

class AlexNetWeights(object):
    """ Struct helper for storing weights """
    def __init__(self):
        pass

def conv(input, kernel, biases, k_h, k_w, c_o, s_h, s_w,  padding="VALID", group=1):
    """
    Convolution layer helper function
    From https://github.com/ethereon/caffe-tensorflow
    """
    c_i = input.get_shape()[-1]
    assert c_i%group==0
    assert c_o%group==0
    convolve = lambda i, k: tf.nn.conv2d(i, k, [1, s_h, s_w, 1], padding=padding)
    
    if group==1:
        conv = convolve(input, kernel)
    else:
        input_groups = tf.split(input, group, axis=3)
        kernel_groups = tf.split(kernel, group, axis=3)
        output_groups = [convolve(i, k) for i,k in zip(input_groups, kernel_groups)]
        conv = tf.concat(output_groups, axis=3)
    return  tf.reshape(tf.nn.bias_add(conv, biases), [-1]+conv.get_shape().as_list()[1:])

class AlexNet(object):
    """ Wrapper for tensorflow AlexNet. Note: training not yet supported.

    Parameters
    ----------
    config : :obj:`autolab_core.YamlConfig`
        specifies the parameters of the network

    Notes
    -----
    Required configuration paramters are specified in Other Parameters

    Other Parameters
    ----------------
    batch_size : int
        size of batches, less than largest possible prediction to save memory
    im_height : int
        height of input images
    im_width : int
        width of input images
    channels : int
        number of channels of input image (should be 3)
    output_layer : :obj:`str`
        name of output layer for classification
    feature_layer : :obj`str`
        name of layer to use for feature extraction (e.g. conv5)
    """
    def __init__(self, config, model_dir=None, use_default_weights=False,
                 dynamic_load=True):
        self._model_dir = model_dir
        self._sess = None
        self._initialized = False
        self._dynamic_load = dynamic_load
        self._parse_config(config)
        if use_default_weights:
            self._initialize()
        elif not self._dynamic_load:
            self._load()

    def _parse_config(self, config):
        """ Parses a tensorflow configuration """
        self._batch_size = config['batch_size']
        self._im_height = config['im_height']
        self._im_width = config['im_width']
        self._num_channels = config['channels']
        self._output_layer = config['out_layer']
        self._feature_layer = config['feature_layer']
        self._out_size = None
        if 'out_size' in config.keys():
            self._out_size = config['out_size']
        self._input_arr = np.zeros([self._batch_size, self._im_height,
                                    self._im_width, self._num_channels])

        if self._model_dir is None:
            self._net_data = np.load(config['caffe_weights']).item()
            self._mean = np.load(config['mean_file'])
            self._model_filename = None
        else:
            self._net_data = None
            self._mean = np.load(os.path.join(self._model_dir, 'mean.npy'))
            self._model_filename = os.path.join(self._model_dir, 'model.ckpt')

    def _load(self):
        """ Loads a model into weights """
        if self._model_filename is None:
            raise ValueError('Model filename not specified')

        # read the input image
        self._graph = tf.Graph()
        with self._graph.as_default():
            # read in filenames
            reader = tf.train.NewCheckpointReader(self._model_filename)

            # load AlexNet weights
            weights = AlexNetWeights()
            weights.conv1W = tf.Variable(reader.get_tensor("Variable")) 
            weights.conv1b = tf.Variable(reader.get_tensor("Variable_1"))
            weights.conv2W = tf.Variable(reader.get_tensor("Variable_2"))
            weights.conv2b = tf.Variable(reader.get_tensor("Variable_3"))
            weights.conv3W = tf.Variable(reader.get_tensor("Variable_4"))
            weights.conv3b = tf.Variable(reader.get_tensor("Variable_5"))
            weights.conv4W = tf.Variable(reader.get_tensor("Variable_6"))
            weights.conv4b = tf.Variable(reader.get_tensor("Variable_7"))
            weights.conv5W = tf.Variable(reader.get_tensor("Variable_8"))
            weights.conv5b = tf.Variable(reader.get_tensor("Variable_9"))
            weights.fc6W = tf.Variable(reader.get_tensor("Variable_10"))
            weights.fc6b = tf.Variable(reader.get_tensor("Variable_11"))
            weights.fc7W = tf.Variable(reader.get_tensor("Variable_12"))
            weights.fc7b = tf.Variable(reader.get_tensor("Variable_13"))
            weights.fc8W = tf.Variable(reader.get_tensor("Variable_14"))
            weights.fc8b = tf.Variable(reader.get_tensor("Variable_15"))

            # form network
            self._input_node = tf.placeholder(tf.float32, (self._batch_size, self._im_height, self._im_width, self._num_channels))
            self._output_tensor = self.build_alexnet(weights)
            self._feature_tensor = self.build_alexnet(weights, output_layer=self._feature_layer)
            self._initialized = True

    def _initialize(self):
        """ Open from caffe weights """
        self._graph = tf.Graph()
        with self._graph.as_default():
            self._input_node = tf.placeholder(tf.float32, (self._batch_size, self._im_height, self._im_width, self._num_channels))
            weights = self.build_alexnet_weights()
            self._output_tensor = self.build_alexnet(weights)
            self._feature_tensor = self.build_alexnet(weights, output_layer=self._feature_layer)
            self._initialized = True

    def open_session(self):
        """ Open tensorflow session. Exposed for memory management. """
        with self._graph.as_default():
            init = tf.initialize_all_variables()
            self._sess = tf.Session()
            self._sess.run(init)

    def close_session(self):
        """ Close tensorflow session. Exposes for memory management. """
        with self._graph.as_default():
            self._sess.close()
            self._sess = None

    def predict(self, image_arr, featurize=False):
        """ Predict a set of images in batches.

        Parameters
        ----------
        image_arr : NxHxWxC :obj:`numpy.ndarray`
            input set of images in a num_images x image height x image width x image channels array (must match parameters of network)
        featurize : bool
            whether or not to use the featurization layer or classification output layer

        Returns
        -------
        :obj:`numpy.ndarray`
            num_images x feature_dim containing the output values for each input image
        """
        # setup prediction
        num_images = image_arr.shape[0]
        output_arr = None

        # predict by filling in image array in batches
        close_sess = False
        if not self._initialized and self._dynamic_load:
            self._load()
        with self._graph.as_default():
            if self._sess is None: 
                close_sess = True               
                self.open_session()

            i = 0
            while i < num_images:
                dim = min(self._batch_size, num_images-i)
                cur_ind = i
                end_ind = cur_ind + dim
                self._input_arr[:dim,:,:,:] = image_arr[cur_ind:end_ind,:,:,:] - self._mean
                if featurize:
                    output = self._sess.run(self._feature_tensor,
                                            feed_dict={self._input_node: self._input_arr})
                else:
                    output = self._sess.run(self._output_tensor,
                                            feed_dict={self._input_node: self._input_arr})
                if output_arr is None:
                    output_arr = output
                else:
                    output_arr = np.r_[output_arr, output]

                i = end_ind
            if close_sess:
                self.close_session()
        return output_arr[:num_images,...]

    def featurize(self, image_arr):
        """ Featurize a set of images in batches.

        Parameters
        ----------
        image_arr : NxHxWxC :obj:`numpy.ndarray`
            input set of images in a num_images x image height x image width x image channels array (must match parameters of network)

        Returns
        -------
        :obj:`numpy.ndarray`
            num_images x feature_dim containing the output values for each input image
        """
        return self.predict(image_arr, featurize=True)

    def build_alexnet_weights(self):
        """ Build a set of convnet weights for AlexNet """
        net_data = self._net_data
        #conv1
        #conv(11, 11, 96, 4, 4, padding='VALID', name='conv1')
        k_h = 11; k_w = 11; c_o = 96; s_h = 4; s_w = 4
        conv1W = tf.Variable(net_data["conv1"][0])
        conv1b = tf.Variable(net_data["conv1"][1])
  
        #conv2
        #conv(5, 5, 256, 1, 1, group=2, name='conv2')
        k_h = 5; k_w = 5; c_o = 256; s_h = 1; s_w = 1; group = 2
        conv2W = tf.Variable(net_data["conv2"][0])
        conv2b = tf.Variable(net_data["conv2"][1])
        
        #conv3
        #conv(3, 3, 384, 1, 1, name='conv3')
        k_h = 3; k_w = 3; c_o = 384; s_h = 1; s_w = 1; group = 1
        conv3W = tf.Variable(net_data["conv3"][0])
        conv3b = tf.Variable(net_data["conv3"][1])
        
        #conv4
        #conv(3, 3, 384, 1, 1, group=2, name='conv4')
        k_h = 3; k_w = 3; c_o = 384; s_h = 1; s_w = 1; group = 2
        conv4W = tf.Variable(net_data["conv4"][0])
        conv4b = tf.Variable(net_data["conv4"][1])
        
        #conv5
        #conv(3, 3, 256, 1, 1, group=2, name='conv5')
        k_h = 3; k_w = 3; c_o = 256; s_h = 1; s_w = 1; group = 2
        conv5W = tf.Variable(net_data["conv5"][0])
        conv5b = tf.Variable(net_data["conv5"][1])
    
        #fc6
        #fc(4096, name='fc6')
        fc6_in_size = net_data["fc6"][0].shape[0]
        fc6_out_size = net_data["fc6"][0].shape[1]
        fc6W = tf.Variable(net_data["fc6"][0])
        fc6b = tf.Variable(net_data["fc6"][1])

        #fc7
        #fc(4096, name='fc7') 
        fc7_in_size = fc6_out_size
        fc7_out_size = net_data["fc7"][0].shape[1]
        fc7W = tf.Variable(net_data["fc7"][0])
        fc7b = tf.Variable(net_data["fc7"][1])
    
        #fc8
        #fc(num_cats, relu=False, name='fc8')                       
        fc8_in_size = fc7_out_size
        fc8_out_size = self._out_size
        fc8W = tf.Variable(tf.truncated_normal([fc8_in_size, fc8_out_size],
                                               stddev=0.01,
                                               seed=None))
        fc8b = tf.Variable(tf.constant(0.0, shape=[fc8_out_size]))
  
        # make return object
        weights = AlexNetWeights()
        weights.conv1W = conv1W
        weights.conv1b = conv1b
        weights.conv2W = conv2W
        weights.conv2b = conv2b
        weights.conv3W = conv3W
        weights.conv3b = conv3b 
        weights.conv4W = conv4W
        weights.conv4b = conv4b 
        weights.conv5W = conv5W
        weights.conv5b = conv5b 
        weights.fc6W = fc6W
        weights.fc6b = fc6b
        weights.fc7W = fc7W
        weights.fc7b = fc7b
        weights.fc8W = fc8W
        weights.fc8b = fc8b
        return weights

    def build_alexnet(self, weights, output_layer=None):
        """ Connects graph of alexnet from weights """
        if output_layer is None:
            output_layer = self._output_layer

        #conv1
        #conv(11, 11, 96, 4, 4, padding='VALID', name='conv1')
        k_h = 11; k_w = 11; c_o = 96; s_h = 4; s_w = 4
        conv1_in = conv(self._input_node, weights.conv1W, weights.conv1b, k_h, k_w, c_o, s_h, s_w, padding="SAME", group=1)
        conv1 = tf.nn.relu(conv1_in)
        if output_layer == 'conv1':
            return tf.contrib.layers.flatten(conv1)
        
        #lrn1
        #lrn(2, 2e-05, 0.75, name='norm1')
        radius = 2; alpha = 2e-05; beta = 0.75; bias = 1.0
        lrn1 = tf.nn.local_response_normalization(conv1,
                                                  depth_radius=radius,
                                                  alpha=alpha,
                                                  beta=beta,
                                                  bias=bias)
        
        #maxpool1
        #max_pool(3, 3, 2, 2, padding='VALID', name='pool1')
        k_h = 3; k_w = 3; s_h = 2; s_w = 2; padding = 'VALID'
        maxpool1 = tf.nn.max_pool(lrn1, ksize=[1, k_h, k_w, 1], strides=[1, s_h, s_w, 1], padding=padding)
        
        
        #conv2
        #conv(5, 5, 256, 1, 1, group=2, name='conv2')
        k_h = 5; k_w = 5; c_o = 256; s_h = 1; s_w = 1; group = 2
        conv2_in = conv(maxpool1, weights.conv2W, weights.conv2b, k_h, k_w, c_o, s_h, s_w, padding="SAME", group=group)
        conv2 = tf.nn.relu(conv2_in)
        if output_layer == 'conv2':
            return tf.contrib.layers.flatten(conv2)
                
        #lrn2
        #lrn(2, 2e-05, 0.75, name='norm2')
        radius = 2; alpha = 2e-05; beta = 0.75; bias = 1.0
        lrn2 = tf.nn.local_response_normalization(conv2,
                                                  depth_radius=radius,
                                                  alpha=alpha,
                                                  beta=beta,
                                                  bias=bias)
        
        #maxpool2
        #max_pool(3, 3, 2, 2, padding='VALID', name='pool2')                                                  
        k_h = 3; k_w = 3; s_h = 2; s_w = 2; padding = 'VALID'
        maxpool2 = tf.nn.max_pool(lrn2, ksize=[1, k_h, k_w, 1], strides=[1, s_h, s_w, 1], padding=padding)
        
        #conv3
        #conv(3, 3, 384, 1, 1, name='conv3')
        k_h = 3; k_w = 3; c_o = 384; s_h = 1; s_w = 1; group = 1
        conv3_in = conv(maxpool2, weights.conv3W, weights.conv3b, k_h, k_w, c_o, s_h, s_w, padding="SAME", group=group)
        conv3 = tf.nn.relu(conv3_in)
        if output_layer == 'conv3':
            return tf.contrib.layers.flatten(conv3)
        
        #conv4
        #conv(3, 3, 384, 1, 1, group=2, name='conv4')
        k_h = 3; k_w = 3; c_o = 384; s_h = 1; s_w = 1; group = 2
        conv4_in = conv(conv3, weights.conv4W, weights.conv4b, k_h, k_w, c_o, s_h, s_w, padding="SAME", group=group)
        conv4 = tf.nn.relu(conv4_in)
        if output_layer == 'conv4':
            return tf.contrib.layers.flatten(conv4)
        
        #conv5
        #conv(3, 3, 256, 1, 1, group=2, name='conv5')
        k_h = 3; k_w = 3; c_o = 256; s_h = 1; s_w = 1; group = 2
        conv5_in = conv(conv4, weights.conv5W, weights.conv5b, k_h, k_w, c_o, s_h, s_w, padding="SAME", group=group)
        conv5 = tf.nn.relu(conv5_in)
        if output_layer == 'conv5':
            return tf.contrib.layers.flatten(conv5)
        
        #maxpool5
        #max_pool(3, 3, 2, 2, padding='VALID', name='pool5')
        k_h = 3; k_w = 3; s_h = 2; s_w = 2; padding = 'VALID'
        maxpool5 = tf.nn.max_pool(conv5, ksize=[1, k_h, k_w, 1], strides=[1, s_h, s_w, 1], padding=padding)

        #fc6
        #fc(4096, name='fc6')
        fc6 = tf.nn.relu_layer(tf.reshape(maxpool5, [-1, int(np.prod(maxpool5.get_shape()[1:]))]), weights.fc6W, weights.fc6b)
        if output_layer == 'fc6':
            return fc6
            
        #fc7
        #fc(4096, name='fc7') 
        fc7 = tf.nn.relu_layer(fc6, weights.fc7W, weights.fc7b)
        if output_layer == 'fc7':
            return fc7
    
        #fc8
        #fc(num_cats, relu=False, name='fc8')
        fc8 = tf.nn.xw_plus_b(fc7, weights.fc8W, weights.fc8b)
        if output_layer == 'fc8':
            return fc8        

        #softmax
        sm = tf.nn.softmax(fc8)
        return sm




'''
Abstraction for interacting with video devices that interface with OpenCV
Author: Jacky Liang
'''
import cv2
import numpy as np
from time import time

from .camera_sensor import CameraSensor
from .image import ColorImage

class OpenCVCameraSensor(CameraSensor):

    def __init__(self, device_id, upside_down=False):
        self._device_id = device_id
        self._upside_down = upside_down

    def start(self):
        """ Starts the OpenCVCameraSensor Stream
        Raises:
            Exception if unable to open stream
        """
        self._sensor = cv2.VideoCapture(self._device_id)
        if not self._sensor.isOpened():
            raise Exception("Unable to open OpenCVCameraSensor for id {0}".format(self._device_id))
        self.flush()

    def flush(self):
        for _ in range(6):
            self._sensor.read()

    def stop(self):
        """ Stops the OpenCVCameraSensor Stream """
        self._sensor.release()

    def frames(self, flush=True):
        """ Returns the latest color image from the stream
        Raises:
            Exception if opencv sensor gives ret_val of 0
        """
        self.flush()
        ret_val, frame = self._sensor.read()
        if not ret_val:
            raise Exception("Unable to retrieve frame from OpenCVCameraSensor for id {0}".format(self._device_id))
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        if self._upside_down:
            frame = np.flipud(frame).astype(np.uint8)
            frame = np.fliplr(frame).astype(np.uint8)
        return ColorImage(frame)



"""
Encapsulates camera intrinsic parameters for projecting / deprojecitng points
Author: Jeff Mahler
"""
import copy
import logging
import numpy as np
import json
import os

from autolab_core import Point, PointCloud, ImageCoords

from .constants import INTR_EXTENSION
from .image import DepthImage, PointCloudImage

class OrthographicIntrinsics(object):
    """A set of intrinsic parameters for orthographic point cloud projections
    """

    def __init__(self, frame,
                 vol_width,
                 vol_height,
                 vol_depth,
                 plane_height,
                 plane_width,
                 depth_scale=1.0):
        """Initialize a CameraIntrinsics model.

        Parameters
        ----------
        frame : :obj:`str`
            The frame of reference for the point cloud.
        vol_height : float
            The height of the 3D projection volume in meters.
        vol_width : float
            The width of the 3D projection volume in meters.
        vol_depth : float
            The depth of the 3D projection volume in meters.
        plane_height : float
            The height of the projection plane in pixels.
        plane_width : float
            The width of the projection plane in pixels.
        depth_scale : float
            The scale of the depth values.
        """
        self._frame = frame
        self._vol_height = float(vol_height)
        self._vol_width = float(vol_width)
        self._vol_depth = float(vol_depth)
        self._plane_height = float(plane_height)
        self._plane_width = float(plane_width)
        self._depth_scale = float(depth_scale)

    @property
    def frame(self):
        """:obj:`str` : The frame of reference for the point cloud.
        """
        return self._frame

    @property
    def plane_height(self):
        """float : The height of the projection plane in pixels.
        """
        return self._height

    @property
    def plane_width(self):
        """float : The width of the projection plane in pixels.
        """
        return self._plane_width

    @property
    def S(self):
        """:obj:`numpy.ndarray` : The 3x3 scaling matrix for this projection
        """
        S = np.array([[self._plane_width / self._vol_width, 0, 0],
                      [0, self._plane_height / self._vol_height, 0],
                      [0, 0, self._depth_scale / self._vol_depth]])
        return S

    @property
    def t(self):
        """:obj:`numpy.ndarray` : The 3x1 translation matrix for this projection
        """
        t = np.array([self._plane_width / 2,
                      self._plane_height / 2,
                      self._depth_scale / 2])
        return t
    
    @property
    def proj_matrix(self):
        """:obj:`numpy.ndarray` : The 4x4 projection matrix for this camera.
        """
        return self.P

    @property
    def P(self):
        """:obj:`numpy.ndarray` : The 4x4 projection matrix for this camera.
        """
        P = np.r_[np.c_[self.S, self.t], np.array([0,0,0,1])]
        return P

    def project(self, point_cloud, round_px=True):
        """Projects a point cloud onto the projection plane.

        Parameters
        ----------
        point_cloud : :obj:`autolab_core.PointCloud` or :obj:`autolab_core.Point`
            A PointCloud or Point to project onto the camera image plane.

        round_px : bool
            If True, projections are rounded to the nearest pixel.

        Returns
        -------
        :obj:`autolab_core.ImageCoords` or :obj:`autolab_core.Point`
            A corresponding set of image coordinates representing the given
            PointCloud's projections onto the camera image plane. If the input
            was a single Point, returns a 2D Point in the camera plane.

        Raises
        ------
        ValueError
            If the input is not a PointCloud or Point in the same reference
            frame as the camera.
        """
        if not isinstance(point_cloud, PointCloud) and not (isinstance(point_cloud, Point) and point_cloud.dim == 3):
            raise ValueError('Must provide PointCloud or 3D Point object for projection')
        if point_cloud.frame != self._frame:
            raise ValueError('Cannot project points in frame %s into camera with frame %s' %(point_cloud.frame, self._frame))

        points_proj = self.S.dot(point_cloud.data) + self.t
        if len(points_proj.shape) == 1:
            points_proj = points_proj[:, np.newaxis]
        point_depths = np.tile(points_proj[2,:], [3, 1])
        points_proj = np.divide(points_proj, point_depths)
        if round_px:
            points_proj = np.round(points_proj)

        if isinstance(point_cloud, Point):
            return Point(data=points_proj[:2,:].astype(np.int16), frame=self._frame)
        return ImageCoords(data=points_proj[:2,:].astype(np.int16), frame=self._frame)

    def project_to_image(self, point_cloud, round_px=True):
        """Projects a point cloud onto the camera image plane and creates
        a depth image. Zero depth means no point projected into the camera
        at that pixel location (i.e. infinite depth).

        Parameters
        ----------
        point_cloud : :obj:`autolab_core.PointCloud` or :obj:`autolab_core.Point`
            A PointCloud or Point to project onto the camera image plane.

        round_px : bool
            If True, projections are rounded to the nearest pixel.

        Returns
        -------
        :obj:`DepthImage`
            A DepthImage generated from projecting the point cloud into the
            camera.

        Raises
        ------
        ValueError
            If the input is not a PointCloud or Point in the same reference
            frame as the camera.
        """
        if not isinstance(point_cloud, PointCloud) and not (isinstance(point_cloud, Point) and point_cloud.dim == 3):
            raise ValueError('Must provide PointCloud or 3D Point object for projection')
        if point_cloud.frame != self._frame:
            raise ValueError('Cannot project points in frame %s into camera with frame %s' %(point_cloud.frame, self._frame))

        points_proj = self.S.dot(point_cloud.data) + self.t
        if len(points_proj.shape) == 1:
            points_proj = points_proj[:, np.newaxis]
        point_depths = points_proj[2,:]
        point_z = np.tile(point_depths, [3, 1])
        points_proj = np.divide(points_proj, point_z)
        if round_px:
            points_proj = np.round(points_proj)
        points_proj = points_proj[:2,:].astype(np.int16)

        valid_ind = np.where((points_proj[0,:] >= 0) & \
                             (points_proj[1,:] >= 0) & \
                             (points_proj[0,:] < self.width) & \
                             (points_proj[1,:] < self.height))[0]

        depth_data = np.zeros([self.height, self.width])
        depth_data[points_proj[1,valid_ind], points_proj[0,valid_ind]] = point_depths[valid_ind]
        return DepthImage(depth_data, frame=self.frame)

    def deproject(self, depth_image):
        """Deprojects a DepthImage into a PointCloud.

        Parameters
        ----------
        depth_image : :obj:`DepthImage`
            The 2D depth image to projet into a point cloud.

        Returns
        -------
        :obj:`autolab_core.PointCloud`
            A 3D point cloud created from the depth image.

        Raises
        ------
        ValueError
            If depth_image is not a valid DepthImage in the same reference frame
            as the camera.
        """
        # check valid input
        if not isinstance(depth_image, DepthImage):
            raise ValueError('Must provide DepthImage object for projection')
        if depth_image.frame != self._frame:
            raise ValueError('Cannot deproject points in frame %s from camera with frame %s' %(depth_image.frame, self._frame))

        # create homogeneous pixels
        row_indices = np.arange(depth_image.height)
        col_indices = np.arange(depth_image.width)
        pixel_grid = np.meshgrid(col_indices, row_indices)
        pixels = np.c_[pixel_grid[0].flatten(), pixel_grid[1].flatten()].T
        depth_data = depth_image.data.flatten()
        pixels_homog = np.r_[pixels, depth_data.reshape(1, depth_data.shape[0])]

        # deproject
        points_3d = np.linalg.inv(self.S).dot(pixels_homog - np.tile(self.t.reshape(3,1), [1, pixels_homog.shape[1]]))
        return PointCloud(data=points_3d, frame=self._frame)

    def deproject_to_image(self, depth_image):
        """Deprojects a DepthImage into a PointCloudImage.

        Parameters
        ----------
        depth_image : :obj:`DepthImage`
            The 2D depth image to projet into a point cloud.

        Returns
        -------
        :obj:`PointCloudImage`
            A point cloud image created from the depth image.

        Raises
        ------
        ValueError
            If depth_image is not a valid DepthImage in the same reference frame
            as the camera.
        """
        point_cloud = self.deproject(depth_image)
        point_cloud_im_data = point_cloud.data.T.reshape(depth_image.height, depth_image.width, 3)
        return PointCloudImage(data=point_cloud_im_data,
                               frame=self._frame)

    def deproject_pixel(self, depth, pixel):
        """Deprojects a single pixel with a given depth into a 3D point.

        Parameters
        ----------
        depth : float
            The depth value at the given pixel location.

        pixel : :obj:`autolab_core.Point`
            A 2D point representing the pixel's location in the camera image.

        Returns
        -------
        :obj:`autolab_core.Point`
            The projected 3D point.

        Raises
        ------
        ValueError
            If pixel is not a valid 2D Point in the same reference frame
            as the camera.
        """
        if not isinstance(pixel, Point) and not pixel.dim == 2:
            raise ValueError('Must provide 2D Point object for pixel projection')
        if pixel.frame != self._frame:
            raise ValueError('Cannot deproject pixel in frame %s from camera with frame %s' %(pixel.frame, self._frame))

        point = np.r_[pixel.data, depth]
        point_3d = np.linalg.inv(self.S).dot(point - self.t)
        return Point(data=point_3d, frame=self._frame)

    def save(self, filename):
        """Save the CameraIntrinsics object to a .intr file.

        Parameters
        ----------
        filename : :obj:`str`
            The .intr file to save the object to.

        Raises
        ------
        ValueError
            If filename does not have the .intr extension.
        """
        file_root, file_ext = os.path.splitext(filename)
        if file_ext.lower() != INTR_EXTENSION:
            raise ValueError('Extension %s not supported for OrhtographicIntrinsics. Must be stored with extension %s' %(file_ext, INTR_EXTENSION))

        camera_intr_dict = copy.deepcopy(self.__dict__)
        f = open(filename, 'w')
        json.dump(camera_intr_dict, f)
        f.close()

    @staticmethod
    def load(filename):
        """Load a CameraIntrinsics object from a file.

        Parameters
        ----------
        filename : :obj:`str`
            The .intr file to load the object from.

        Returns
        -------
        :obj:`CameraIntrinsics`
            The CameraIntrinsics object loaded from the file.

        Raises
        ------
        ValueError
            If filename does not have the .intr extension.
        """
        file_root, file_ext = os.path.splitext(filename)
        if file_ext.lower() != INTR_EXTENSION:
            raise ValueError('Extension %s not supported for CameraIntrinsics. Must be stored with extension %s' %(file_ext, INTR_EXTENSION))

        f = open(filename, 'r')
        ci = json.load(f)
        f.close()
        return OrthographicIntrinsics(frame=ci['_frame'],
                                      vol_height=ci['_vol_height'],
                                      vol_width=ci['_vol_width'],
                                      vol_depth=ci['_vol_depth'],
                                      plane_height=ci['_plane_height'],
                                      plane_width=ci['_plane_width'],
                                      depth_scale=ci['_depth_scale'])



"""
RGBD Sensor factory
Author: Jeff Mahler
"""
from . import Kinect2Sensor, PrimesenseSensor, RealSenseSensor, VirtualSensor, PrimesenseSensor_ROS, EnsensoSensor, PhoXiSensor, TensorDatasetVirtualSensor, KinectSensorBridged, WebcamSensor, ColorizedPhoXiSensor

class RgbdSensorFactory:
    """ Factory class for Rgbd camera sensors. """

    @staticmethod
    def sensor(sensor_type, cfg):
        """ Creates a camera sensor of the specified type.

        Parameters
        ----------
        sensor_type : :obj:`str`
            the type of the sensor (real or virtual)
        cfg : :obj:`YamlConfig`
            dictionary of parameters for sensor initialization
        """
        sensor_type = sensor_type.lower()
        if sensor_type == 'kinect2':
            s = Kinect2Sensor(packet_pipeline_mode=cfg['pipeline_mode'],
                              device_num=cfg['device_num'],
                              frame=cfg['frame'])
        elif sensor_type == 'bridged_kinect2':
            s = KinectSensorBridged(quality=cfg['quality'], frame=cfg['frame'])
        elif sensor_type == 'primesense':
            flip_images = True
            if 'flip_images' in cfg.keys():
                flip_images = cfg['flip_images']
            s = PrimesenseSensor(auto_white_balance=cfg['auto_white_balance'],
                                 flip_images=flip_images,
                                 frame=cfg['frame'])
        elif sensor_type == 'virtual':
            s = VirtualSensor(cfg['image_dir'],
                              frame=cfg['frame'])
        elif sensor_type == 'tensor_dataset':
            s = TensorDatasetVirtualSensor(cfg['dataset_dir'],
                                           frame=cfg['frame'])
        elif sensor_type == 'primesense_ros':
            s = PrimesenseSensor_ROS(frame=cfg['frame'])
        elif sensor_type == 'ensenso':
            s = EnsensoSensor(frame=cfg['frame'])
        elif sensor_type == 'phoxi':
            s = PhoXiSensor(frame=cfg['frame'],
                            device_name=cfg['device_name'],
                            size=cfg['size'])
        elif sensor_type == 'webcam':
            s = WebcamSensor(frame=cfg['frame'],
                             device_id=cfg['device_id'])
        elif sensor_type == 'colorized_phoxi':
            s = ColorizedPhoXiSensor(frame=cfg['frame'], phoxi_config=cfg['phoxi_config'],
                                     webcam_config=cfg['webcam_config'], calib_dir=cfg['calib_dir'])
        elif sensor_type == 'realsense':
            s = RealSenseSensor(
                cam_id=cfg['cam_id'],
                filter_depth=cfg['filter_depth'],
                frame=cfg['frame'],
            )
        else:
            raise ValueError('RGBD sensor type %s not supported' %(sensor_type))
        return s



"""
Setup of meshpy python codebase
Author: Jeff Mahler
"""
from setuptools import setup

requirements = [
    'numpy',
    'scipy',
    'autolab_core',
    'matplotlib',
    'multiprocess',
    'opencv-python',
    'keras',
    'cycler',
    'Pillow',
    'pyserial>=3.4',
    'ipython==5.5.0',
    'scikit-image',
    'scikit-learn',
    'scikit-video'
]

exec(open('perception/version.py').read())

setup(name='autolab_perception',
      version=__version__,
      description='Perception utilities for the Berkeley AutoLab',
      author='Jeff Mahler',
      author_email='jmahler@berkeley.edu',
      license = 'Apache Software License',
      url = 'https://github.com/BerkeleyAutomation/perception',
      keywords = 'robotics grasping vision perception',
      classifiers = [
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Natural Language :: English',
          'Topic :: Scientific/Engineering'
      ],
      packages=['perception'],
      install_requires = requirements,
      extras_require = { 'docs' : [
                            'sphinx',
                            'sphinxcontrib-napoleon',
                            'sphinx_rtd_theme'
                        ],
                       'ros' : [
                           'primesense',
                           'rospkg',
                           'catkin_pkg',
                           'empy'
                        ],
    }
)



"""
Classes for point set registration using variants of Iterated-Closest Point
Author: Jeff Mahler
"""
from abc import ABCMeta, abstractmethod
import logging
import numpy as np

from autolab_core import RigidTransform, PointCloud, NormalCloud, skew

from .feature_matcher import PointToPlaneFeatureMatcher

class RegistrationResult(object):
    """ Struct to hold results of point set registration.

    Attributes
    ----------
    T_source_target : :obj:`autolab_core.RigidTranform`
        transformation from source to target frame
    cost : float
        numeric value of the registration objective for the given transform
    """
    def __init__(self, T_source_target, cost):
        self.T_source_target = T_source_target
        self.cost = cost

class IterativeRegistrationSolver:
    """ Abstract class for iterative registration solvers. """
    __metaclass__ = ABCMeta

    @abstractmethod
    def register(self, source_point_cloud, target_point_cloud,
                 source_normal_cloud, target_normal_cloud, matcher,
                 num_iterations=1, compute_total_cost=True, match_centroids=False,
                 vis=False):
        """ Iteratively register objects to one another.

        Parameters
        ----------
        source_point_cloud : :obj:`autolab_core.PointCloud`
            source object points
        target_point_cloud : :obj`autolab_core.PointCloud`
            target object points
        source_normal_cloud : :obj:`autolab_core.NormalCloud`
            source object outward-pointing normals
        target_normal_cloud : :obj:`autolab_core.NormalCloud`
            target object outward-pointing normals
        matcher : :obj:`PointToPlaneFeatureMatcher`
            object to match the point sets
        num_iterations : int
            the number of iterations to run
        compute_total_cost : bool
            whether or not to compute the total cost upon termination.
        match_centroids : bool
            whether or not to match the centroids of the point clouds
        
        Returns
        -------
        :obj`RegistrationResult`
            results containing source to target transformation and cost
        """
        pass

class PointToPlaneICPSolver(IterativeRegistrationSolver):
    """ Performs Iterated Closest Point with an objective weighted between point-to-point and point-to-plane.

    Attributes
    ----------
    sample_size : int
        number of randomly sampled points to use per iteration
    cost_sample_size : int
        number of randomly sampled points to use for cost evaluations
    gamma : float
        weight of point-to-point objective relative to point-to-plane objective
    mu : float
        regularizer for matrix inversion in the Gauss-Newton step
    """
    def __init__(self, sample_size=100, cost_sample_size=100, gamma=100.0, mu=1e-2):
        self.sample_size_ = sample_size
        self.cost_sample_size_ = cost_sample_size
        self.gamma_ = gamma
        self.mu_ = mu
        IterativeRegistrationSolver.__init__(self)
    
    def register(self, source_point_cloud, target_point_cloud,
                 source_normal_cloud, target_normal_cloud, matcher,
                 num_iterations=1, compute_total_cost=True, match_centroids=False,
                 vis=False):
        """
        Iteratively register objects to one another using a modified version of point to plane ICP.
        The cost func is PointToPlane_COST + gamma * PointToPoint_COST.
        Uses a `stochastic Gauss-Newton step` where on each iteration a smaller number of points is sampled.

        Parameters
        ----------
        source_point_cloud : :obj:`autolab_core.PointCloud`
            source object points
        target_point_cloud : :obj`autolab_core.PointCloud`
            target object points
        source_normal_cloud : :obj:`autolab_core.NormalCloud`
            source object outward-pointing normals
        target_normal_cloud : :obj:`autolab_core.NormalCloud`
            target object outward-pointing normals
        matcher : :obj:`PointToPlaneFeatureMatcher`
            object to match the point sets
        num_iterations : int
            the number of iterations to run
        compute_total_cost : bool
            whether or not to compute the total cost upon termination.
        match_centroids : bool
            whether or not to match the centroids of the point clouds
        
        Returns
        -------
        :obj`RegistrationResult`
            results containing source to target transformation and cost
        """
        # check valid data
        if not isinstance(source_point_cloud, PointCloud) or not isinstance(target_point_cloud, PointCloud):
            raise ValueError('Source and target point clouds must be PointCloud objects')
        if not isinstance(source_normal_cloud, NormalCloud) or not isinstance(target_normal_cloud, NormalCloud):
            raise ValueError('Source and target normal clouds must be NormalCloud objects')
        if not isinstance(matcher, PointToPlaneFeatureMatcher):
            raise ValueError('Feature matcher must be a PointToPlaneFeatureMatcher object')
        if source_point_cloud.num_points != source_normal_cloud.num_points or target_point_cloud.num_points != target_normal_cloud.num_points:
            raise ValueError('Input point clouds must have the same number of points as corresponding normal cloud')

        # extract source and target point and normal data arrays
        orig_source_points = source_point_cloud.data.T
        orig_target_points = target_point_cloud.data.T
        orig_source_normals = source_normal_cloud.data.T
        orig_target_normals = target_normal_cloud.data.T

        # setup the problem
        normal_norms = np.linalg.norm(orig_target_normals, axis=1)
        valid_inds = np.nonzero(normal_norms)
        orig_target_points = orig_target_points[valid_inds[0],:]
        orig_target_normals = orig_target_normals[valid_inds[0],:]

        normal_norms = np.linalg.norm(orig_source_normals, axis=1)
        valid_inds = np.nonzero(normal_norms)
        orig_source_points = orig_source_points[valid_inds[0],:]
        orig_source_normals = orig_source_normals[valid_inds[0],:]

        # alloc buffers for solutions
        source_mean_point = np.mean(orig_source_points, axis=0)
        target_mean_point = np.mean(orig_target_points, axis=0)
        R_sol = np.eye(3)
        t_sol = np.zeros([3, 1]) #init with diff between means
        if match_centroids:
            t_sol[:,0] = target_mean_point - source_mean_point

        # iterate through
        for i in range(num_iterations):
            logging.info('Point to plane ICP iteration %d' %(i))

            # subsample points
            source_subsample_inds = np.random.choice(orig_source_points.shape[0], size=self.sample_size_)
            source_points = orig_source_points[source_subsample_inds,:]
            source_normals = orig_source_normals[source_subsample_inds,:]
            target_subsample_inds = np.random.choice(orig_target_points.shape[0], size=self.sample_size_)
            target_points = orig_target_points[target_subsample_inds,:]
            target_normals = orig_target_normals[target_subsample_inds,:]

            # transform source points
            source_points = (R_sol.dot(source_points.T) + np.tile(t_sol, [1, source_points.shape[0]])).T
            source_normals = (R_sol.dot(source_normals.T)).T
        
            # closest points
            corrs = matcher.match(source_points, target_points, source_normals, target_normals)

            # solve optimal rotation + translation
            valid_corrs = np.where(corrs.index_map != -1)[0]
            source_corr_points = corrs.source_points[valid_corrs,:]
            target_corr_points = corrs.target_points[corrs.index_map[valid_corrs], :]
            target_corr_normals = corrs.target_normals[corrs.index_map[valid_corrs], :]

            num_corrs = valid_corrs.shape[0]
            if num_corrs == 0:
                logging.warning('No correspondences found')
                break

            # create A and b matrices for Gauss-Newton step on joint cost function
            A = np.zeros([6,6])
            b = np.zeros([6,1])
            Ap = np.zeros([6,6])
            bp = np.zeros([6,1])
            G = np.zeros([3,6])
            G[:,3:] = np.eye(3)

            for i in range(num_corrs):
                s = source_corr_points[i:i+1,:].T
                t = target_corr_points[i:i+1,:].T
                n = target_corr_normals[i:i+1,:].T
                G[:,:3] = skew(s).T
                A += G.T.dot(n).dot(n.T).dot(G)
                b += G.T.dot(n).dot(n.T).dot(t - s)

                Ap += G.T.dot(G)
                bp += G.T.dot(t - s)
            v = np.linalg.solve(A + self.gamma_*Ap + self.mu_*np.eye(6),
                                b + self.gamma_*bp)

            # create pose values from the solution
            R = np.eye(3)
            R = R + skew(v[:3])
            U, S, V = np.linalg.svd(R)
            R = U.dot(V)
            t = v[3:]

            # incrementally update the final transform
            R_sol = R.dot(R_sol)
            t_sol = R.dot(t_sol) + t

        T_source_target = RigidTransform(R_sol, t_sol, from_frame=source_point_cloud.frame, to_frame=target_point_cloud.frame)

        total_cost = 0
        source_points = (R_sol.dot(orig_source_points.T) + np.tile(t_sol, [1, orig_source_points.shape[0]])).T
        source_normals = (R_sol.dot(orig_source_normals.T)).T

        if compute_total_cost:
            # rematch all points to get the final cost
            corrs = matcher.match(source_points, orig_target_points, source_normals, orig_target_normals)
            valid_corrs = np.where(corrs.index_map != -1)[0]
            num_corrs = valid_corrs.shape[0]
            if num_corrs == 0:
                return RegistrationResult(T_source_target, np.inf)

            # get the corresponding points
            source_corr_points = corrs.source_points[valid_corrs,:]
            target_corr_points = corrs.target_points[corrs.index_map[valid_corrs], :]
            target_corr_normals = corrs.target_normals[corrs.index_map[valid_corrs], :]

            # determine total cost
            source_target_alignment = np.diag((source_corr_points - target_corr_points).dot(target_corr_normals.T))
            point_plane_cost = (1.0 / num_corrs) * np.sum(source_target_alignment * source_target_alignment)
            point_dist_cost = (1.0 / num_corrs) * np.sum(np.linalg.norm(source_corr_points - target_corr_points, axis=1)**2)
            total_cost = point_plane_cost + self.gamma_ * point_dist_cost

        return RegistrationResult(T_source_target, total_cost)

    def register_2d(self, source_point_cloud, target_point_cloud,
                    source_normal_cloud, target_normal_cloud, matcher,
                    num_iterations=1, compute_total_cost=True, vis=False):
        """
        Iteratively register objects to one another using a modified version of point to plane ICP
        which only solves for tx and ty (translation in the plane) and theta (rotation about the z axis).
        The cost func is actually PointToPlane_COST + gamma * PointToPoint_COST
        Points should be specified in the basis of the planar worksurface.

        Parameters
        ----------
        source_point_cloud : :obj:`autolab_core.PointCloud`
            source object points
        target_point_cloud : :obj`autolab_core.PointCloud`
            target object points
        source_normal_cloud : :obj:`autolab_core.NormalCloud`
            source object outward-pointing normals
        target_normal_cloud : :obj:`autolab_core.NormalCloud`
            target object outward-pointing normals
        matcher : :obj:`PointToPlaneFeatureMatcher`
            object to match the point sets
        num_iterations : int
            the number of iterations to run
        compute_total_cost : bool
            whether or not to compute the total cost upon termination.
        
        Returns
        -------
        :obj`RegistrationResult`
            results containing source to target transformation and cost
        """     
        if not isinstance(source_point_cloud, PointCloud) or not isinstance(target_point_cloud, PointCloud):
            raise ValueError('Source and target point clouds must be PointCloud objects')
        if not isinstance(source_normal_cloud, NormalCloud) or not isinstance(target_normal_cloud, NormalCloud):
            raise ValueError('Source and target normal clouds must be NormalCloud objects')
        if not isinstance(matcher, PointToPlaneFeatureMatcher):
            raise ValueError('Feature matcher must be a PointToPlaneFeatureMatcher object')
        if source_point_cloud.num_points != source_normal_cloud.num_points or target_point_cloud.num_points != target_normal_cloud.num_points:
            raise ValueError('Input point clouds must have the same number of points as corresponding normal cloud')

        # extract source and target point and normal data arrays
        orig_source_points = source_point_cloud.data.T
        orig_target_points = target_point_cloud.data.T
        orig_source_normals = source_normal_cloud.data.T
        orig_target_normals = target_normal_cloud.data.T

        # setup the problem
        logging.info('Setting up problem')
        normal_norms = np.linalg.norm(orig_target_normals, axis=1)
        valid_inds = np.nonzero(normal_norms)
        orig_target_points = orig_target_points[valid_inds[0],:]
        orig_target_normals = orig_target_normals[valid_inds[0],:]

        normal_norms = np.linalg.norm(orig_source_normals, axis=1)
        valid_inds = np.nonzero(normal_norms)
        orig_source_points = orig_source_points[valid_inds[0],:]
        orig_source_normals = orig_source_normals[valid_inds[0],:]

        # alloc buffers for solutions
        source_mean_point = np.mean(orig_source_points, axis=0)
        target_mean_point = np.mean(orig_target_points, axis=0)
        R_sol = np.eye(3)
        t_sol = np.zeros([3, 1])

        # iterate through
        for i in range(num_iterations):
            logging.info('Point to plane ICP iteration %d' %(i))

            # subsample points
            source_subsample_inds = np.random.choice(orig_source_points.shape[0], size=self.sample_size_)
            source_points = orig_source_points[source_subsample_inds,:]
            source_normals = orig_source_normals[source_subsample_inds,:]
            target_subsample_inds = np.random.choice(orig_target_points.shape[0], size=self.sample_size_)
            target_points = orig_target_points[target_subsample_inds,:]
            target_normals = orig_target_normals[target_subsample_inds,:]

            # transform source points
            source_points = (R_sol.dot(source_points.T) + np.tile(t_sol, [1, source_points.shape[0]])).T
            source_normals = (R_sol.dot(source_normals.T)).T
        
            # closest points
            corrs = matcher.match(source_points, target_points, source_normals, target_normals)

            # solve optimal rotation + translation
            valid_corrs = np.where(corrs.index_map != -1)[0]
            source_corr_points = corrs.source_points[valid_corrs,:]
            target_corr_points = corrs.target_points[corrs.index_map[valid_corrs], :]
            target_corr_normals = corrs.target_normals[corrs.index_map[valid_corrs], :]

            num_corrs = valid_corrs.shape[0]
            if num_corrs == 0:
                break

            # create A and b matrices for Gauss-Newton step on joint cost function
            A = np.zeros([3,3]) # A and b for point to plane cost
            b = np.zeros([3,1])
            Ap = np.zeros([3,3]) # A and b for point to point cost
            bp = np.zeros([3,1])
            G = np.zeros([3,3])
            G[:2,1:] = np.eye(2)

            for i in range(num_corrs):
                s = source_corr_points[i:i+1,:].T
                t = target_corr_points[i:i+1,:].T
                n = target_corr_normals[i:i+1,:].T
                G[0,0] = -s[1] 
                G[1,0] = s[0]
                A += G.T.dot(n).dot(n.T).dot(G)
                b += G.T.dot(n).dot(n.T).dot(t - s)

                Ap += G.T.dot(G)
                bp += G.T.dot(t - s)
            v = np.linalg.solve(A + self.gamma_*Ap + self.mu_*np.eye(3),
                                b + self.gamma_*bp)

            # create pose values from the solution
            R = np.eye(3)
            R = R + skew(np.array([[0],[0],[v[0,0]]]))
            U, S, V = np.linalg.svd(R)
            R = U.dot(V)
            t = np.array([[v[1,0]], [v[2,0]], [0]])

            # incrementally update the final transform
            R_sol = R.dot(R_sol)
            t_sol = R.dot(t_sol) + t

        # compute solution transform
        T_source_target = RigidTransform(R_sol, t_sol, from_frame=source_point_cloud.frame, to_frame=target_point_cloud.frame)

        total_cost = 0
        if compute_total_cost:
            # subsample points
            source_subsample_inds = np.random.choice(orig_source_points.shape[0], size=self.cost_sample_size_)
            source_points = orig_source_points[source_subsample_inds,:]
            source_normals = orig_source_normals[source_subsample_inds,:]
            target_subsample_inds = np.random.choice(orig_target_points.shape[0], size=self.cost_sample_size_)
            target_points = orig_target_points[target_subsample_inds,:]
            target_normals = orig_target_normals[target_subsample_inds,:]

            # transform source points
            source_points = (R_sol.dot(source_points.T) + np.tile(t_sol, [1, source_points.shape[0]])).T
            source_normals = (R_sol.dot(source_normals.T)).T

            # rematch to get the total cost
            corrs = matcher.match(source_points, target_points, source_normals, target_normals)
            valid_corrs = np.where(corrs.index_map != -1)[0]
            num_corrs = valid_corrs.shape[0]
            if num_corrs == 0:
                return RegistrationResult(T_source_target, np.inf)

            # get the corresponding points
            source_corr_points = corrs.source_points[valid_corrs,:]
            target_corr_points = corrs.target_points[corrs.index_map[valid_corrs], :]
            target_corr_normals = corrs.target_normals[corrs.index_map[valid_corrs], :]

            # determine total cost
            source_target_alignment = np.diag((source_corr_points - target_corr_points).dot(target_corr_normals.T))
            point_plane_cost = (1.0 / num_corrs) * np.sum(source_target_alignment * source_target_alignment)
            point_dist_cost = (1.0 / num_corrs) * np.sum(np.linalg.norm(source_corr_points - target_corr_points, axis=1)**2)
            total_cost = point_plane_cost + self.gamma_ * point_dist_cost

        return RegistrationResult(T_source_target, total_cost)

